#!/usr/bin/env python
from __future__ import annotations

import os
import unittest
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from typing_extensions import assert_type

    from mkdocs.structure.nav import Navigation
else:

    def assert_type(val, typ):
        return None


from mkdocs import plugins
from mkdocs.commands import build
from mkdocs.config import base
from mkdocs.config import config_options as c
from mkdocs.config.base import ValidationError
from mkdocs.exceptions import Abort, BuildError, PluginError
from mkdocs.tests.base import load_config, tempdir


class _DummyPluginConfig(base.Config):
    foo = c.Type(str, default='default foo')
    bar = c.Type(int, default=0)
    dir = c.Optional(c.Dir(exists=False))


class DummyPlugin(plugins.BasePlugin[_DummyPluginConfig]):
    def on_page_content(self, html, **kwargs) -> str:
        """Modify page content by prepending `foo` config value."""
        return f'{self.config.foo} {html}'

    def on_nav(self, nav, **kwargs) -> None:
        """Do nothing (return None) to not modify item."""
        return None

    def on_page_read_source(self, **kwargs) -> str:
        """Create new source by prepending `foo` config value to 'source'."""
        return f'{self.config.foo} source'

    def on_pre_build(self, **kwargs) -> None:
        """Do nothing (return None)."""
        return None


class TestPluginClass(unittest.TestCase):
    def test_valid_plugin_options(self) -> None:
        test_dir = 'test'

        options = {
            'foo': 'some value',
            'dir': test_dir,
        }

        cfg_fname = os.path.join('tmp', 'test', 'fname.yml')
        cfg_fname = os.path.abspath(cfg_fname)

        cfg_dirname = os.path.dirname(cfg_fname)
        expected = {
            'foo': 'some value',
            'bar': 0,
            'dir': os.path.join(cfg_dirname, test_dir),
        }

        plugin = DummyPlugin()
        errors, warnings = plugin.load_config(options, config_file_path=cfg_fname)
        self.assertEqual(errors, [])
        self.assertEqual(warnings, [])

        assert_type(plugin.config, _DummyPluginConfig)
        self.assertEqual(plugin.config, expected)

        assert_type(plugin.config.bar, int)
        self.assertEqual(plugin.config.bar, 0)
        assert_type(plugin.config.dir, Optional[str])

    def test_invalid_plugin_options(self):
        plugin = DummyPlugin()
        errors, warnings = plugin.load_config({'foo': 42})
        self.assertEqual(
            errors,
            [('foo', ValidationError("Expected type: <class 'str'> but received: <class 'int'>"))],
        )
        self.assertEqual(warnings, [])

        errors, warnings = plugin.load_config({'bar': 'a string'})
        self.assertEqual(
            errors,
            [('bar', ValidationError("Expected type: <class 'int'> but received: <class 'str'>"))],
        )
        self.assertEqual(warnings, [])

        errors, warnings = plugin.load_config({'invalid_key': 'value'})
        self.assertEqual(errors, [])
        self.assertEqual(
            warnings, [('invalid_key', "Unrecognised configuration name: invalid_key")]
        )


class TestPluginCollection(unittest.TestCase):
    def test_correct_events_registered(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        collection['foo'] = plugin
        self.assertEqual(
            collection.events,
            {
                'startup': [],
                'shutdown': [],
                'serve': [],
                'config': [],
                'pre_build': [plugin.on_pre_build],
                'files': [],
                'nav': [plugin.on_nav],
                'env': [],
                'post_build': [],
                'build_error': [],
                'pre_template': [],
                'template_context': [],
                'post_template': [],
                'pre_page': [],
                'page_read_source': [plugin.on_page_read_source],
                'page_markdown': [],
                'page_content': [plugin.on_page_content],
                'page_context': [],
                'post_page': [],
            },
        )

    def test_event_priorities(self) -> None:
        class PrioPlugin(plugins.BasePlugin):
            config_scheme = base.get_schema(_DummyPluginConfig)

            @plugins.event_priority(100)
            def on_page_content(self, html, **kwargs) -> None:
                pass

            @plugins.event_priority(50)
            def _on_nav_1(self, nav: Navigation, **kwargs) -> None:
                pass

            @plugins.event_priority(-100)
            def _on_nav_2(self, nav, **kwargs) -> None:
                pass

            on_nav = plugins.CombinedEvent(_on_nav_1, _on_nav_2)

            def on_page_read_source(self, **kwargs) -> None:
                pass

            @plugins.event_priority(-50)
            def on_post_build(self, **kwargs) -> None:
                pass

        collection = plugins.PluginCollection()
        collection['dummy'] = dummy = DummyPlugin()
        with self.assertLogs('mkdocs', level='WARNING') as cm:
            collection['prio'] = prio = PrioPlugin()
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.plugins:Multiple 'on_page_read_source' handlers can't work (both plugins 'dummy' and 'prio' registered one).",
        )

        self.assertEqual(
            collection.events['page_content'],
            [prio.on_page_content, dummy.on_page_content],
        )
        self.assertEqual(
            collection.events['nav'],
            [prio._on_nav_1, dummy.on_nav, prio._on_nav_2],
        )
        self.assertEqual(
            collection.events['page_read_source'],
            [dummy.on_page_read_source, prio.on_page_read_source],
        )
        self.assertEqual(
            collection.events['post_build'],
            [prio.on_post_build],
        )

    def test_set_plugin_on_collection(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        collection['foo'] = plugin
        self.assertEqual(list(collection.items()), [('foo', plugin)])

    def test_set_multiple_plugins_on_collection(self):
        collection = plugins.PluginCollection()
        plugin1 = DummyPlugin()
        collection['foo'] = plugin1
        plugin2 = DummyPlugin()
        with self.assertLogs('mkdocs', level='WARNING') as cm:
            collection['bar'] = plugin2
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.plugins:Multiple 'on_page_read_source' handlers can't work (both plugins 'foo' and 'bar' registered one).",
        )
        self.assertEqual(list(collection.items()), [('foo', plugin1), ('bar', plugin2)])

    def test_run_event_on_collection(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(
            collection.on_page_content('page content', page=None, config={}, files=[]),
            'new page content',
        )

    def test_run_event_twice_on_collection(self):
        collection = plugins.PluginCollection()
        plugin1 = DummyPlugin()
        plugin1.load_config({'foo': 'new'})
        collection['foo'] = plugin1
        plugin2 = DummyPlugin()
        plugin2.load_config({'foo': 'second'})
        with self.assertLogs('mkdocs', level='WARNING') as cm:
            collection['bar'] = plugin2
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.plugins:Multiple 'on_page_read_source' handlers can't work (both plugins 'foo' and 'bar' registered one).",
        )
        self.assertEqual(
            collection.on_page_content('page content', page=None, config={}, files=[]),
            'second new page content',
        )

    def test_event_returns_None(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.on_nav(['nav item'], config={}, files=[]), ['nav item'])

    def test_event_empty_item(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.on_page_read_source(page=None, config={}), 'new source')

    def test_event_empty_item_returns_None(self):
        collection = plugins.PluginCollection()
        plugin = DummyPlugin()
        plugin.load_config({'foo': 'new'})
        collection['foo'] = plugin
        self.assertEqual(collection.on_pre_build(config={}), None)

    def test_run_undefined_event_on_collection(self):
        collection = plugins.PluginCollection()
        self.assertEqual(
            collection.on_page_markdown('page markdown', page=None, config={}, files=[]),
            'page markdown',
        )

    def test_run_unknown_event_on_collection(self):
        collection = plugins.PluginCollection()
        with self.assertRaises(KeyError):
            collection.run_event('unknown', 'page content')

    @tempdir()
    def test_run_build_error_event(self, site_dir):
        build_errors = []

        class PluginRaisingError(plugins.BasePlugin):
            def __init__(self, error_on):
                self.error_on = error_on

            def on_pre_page(self, page, **kwargs):
                if self.error_on == 'pre_page':
                    raise BuildError('pre page error')
                return page

            def on_page_markdown(self, markdown, **kwargs):
                if self.error_on == 'page_markdown':
                    raise BuildError('page markdown error')
                return markdown

            def on_page_content(self, html, **kwargs):
                if self.error_on == 'page_content':
                    raise PluginError('page content error')
                return html

            def on_post_page(self, html, **kwargs):
                if self.error_on == 'post_page':
                    raise ValueError('post page error')

            def on_build_error(self, error, **kwargs):
                build_errors.append(error)

        cfg = load_config(site_dir=site_dir)
        cfg.plugins['errorplugin'] = PluginRaisingError(error_on='pre_page')
        with self.assertLogs('mkdocs', level='ERROR'):
            self.assertRaises(Abort, build.build, cfg)

        cfg = load_config(site_dir=site_dir)
        cfg.plugins['errorplugin'] = PluginRaisingError(error_on='page_markdown')
        with self.assertLogs('mkdocs', level='ERROR'):
            self.assertRaises(Abort, build.build, cfg)

        cfg = load_config(site_dir=site_dir)
        cfg.plugins['errorplugin'] = PluginRaisingError(error_on='page_content')
        with self.assertLogs('mkdocs', level='ERROR'):
            self.assertRaises(Abort, build.build, cfg)

        cfg = load_config(site_dir=site_dir)
        cfg.plugins['errorplugin'] = PluginRaisingError(error_on='post_page')
        with self.assertLogs('mkdocs', level='ERROR'):
            self.assertRaises(ValueError, build.build, cfg)

        cfg = load_config(site_dir=site_dir)
        cfg.plugins['errorplugin'] = PluginRaisingError(error_on='')
        build.build(cfg)

        self.assertEqual(len(build_errors), 4)
        self.assertIs(build_errors[0].__class__, BuildError)
        self.assertEqual(str(build_errors[0]), 'pre page error')
        self.assertIs(build_errors[1].__class__, BuildError)
        self.assertEqual(str(build_errors[1]), 'page markdown error')
        self.assertIs(build_errors[2].__class__, PluginError)
        self.assertEqual(str(build_errors[2]), 'page content error')
        self.assertIs(build_errors[3].__class__, ValueError)
        self.assertEqual(str(build_errors[3]), 'post page error')



import os

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        from babel.messages.frontend import compile_catalog

        for theme in 'mkdocs', 'readthedocs':
            cmd = compile_catalog()
            cmd.directory = os.path.join('mkdocs', 'themes', theme, 'locales')
            cmd.finalize_options()
            cmd.run()



#!/usr/bin/env python

import dataclasses
import datetime
import logging
import os
import posixpath
import stat
import unittest
from unittest import mock

from mkdocs import exceptions, utils
from mkdocs.tests.base import dedent, tempdir
from mkdocs.utils import meta

BASEYML = """
INHERIT: parent.yml
foo: bar
baz:
    sub1: replaced
    sub3: new
deep1:
    deep2-1:
        deep3-1: replaced
"""
PARENTYML = """
foo: foo
baz:
    sub1: 1
    sub2: 2
deep1:
    deep2-1:
        deep3-1: foo
        deep3-2: bar
    deep2-2: baz
"""


class UtilsTests(unittest.TestCase):
    def test_is_markdown_file(self):
        expected_results = {
            'index.md': True,
            'index.markdown': True,
            'index.MARKDOWN': False,
            'index.txt': False,
            'indexmd': False,
        }
        for path, expected_result in expected_results.items():
            with self.subTest(path):
                is_markdown = utils.is_markdown_file(path)
                self.assertEqual(is_markdown, expected_result)

    def test_get_relative_url(self):
        for case in [
            dict(url='foo/bar', other='foo', expected='bar'),
            dict(url='foo/bar.txt', other='foo', expected='bar.txt'),
            dict(url='foo', other='foo/bar', expected='..'),
            dict(url='foo', other='foo/bar.txt', expected='.'),
            dict(url='foo/../../bar', other='.', expected='bar'),
            dict(url='foo/../../bar', other='foo', expected='../bar'),
            dict(url='foo//./bar/baz', other='foo/bar/baz', expected='.'),
            dict(url='a/b/.././../c', other='.', expected='c'),
            dict(url='a/b/c/d/ee', other='a/b/c/d/e', expected='../ee'),
            dict(url='a/b/c/d/ee', other='a/b/z/d/e', expected='../../../c/d/ee'),
            dict(url='foo', other='bar.', expected='foo'),
            dict(url='foo', other='bar./', expected='../foo'),
            dict(url='foo', other='foo/bar./', expected='..'),
            dict(url='foo', other='foo/bar./.', expected='..'),
            dict(url='foo', other='foo/bar././', expected='..'),
            dict(url='foo/', other='foo/bar././', expected='../'),
            dict(url='foo', other='foo', expected='.'),
            dict(url='.foo', other='.foo', expected='.foo'),
            dict(url='.foo/', other='.foo', expected='.foo/'),
            dict(url='.foo', other='.foo/', expected='.'),
            dict(url='.foo/', other='.foo/', expected='./'),
            dict(url='///', other='', expected='./'),
            dict(url='a///', other='', expected='a/'),
            dict(url='a///', other='a', expected='./'),
            dict(url='.', other='here', expected='..'),
            dict(url='..', other='here', expected='..'),
            dict(url='../..', other='here', expected='..'),
            dict(url='../../a', other='here', expected='../a'),
            dict(url='..', other='here.txt', expected='.'),
            dict(url='a', other='', expected='a'),
            dict(url='a', other='..', expected='a'),
            dict(url='a', other='b', expected='../a'),
            # The dots are considered a file. Documenting a long-standing bug:
            dict(url='a', other='b/..', expected='../a'),
            dict(url='a', other='b/../..', expected='a'),
            dict(url='a/..../b', other='a/../b', expected='../a/..../b'),
            dict(url='a/я/b', other='a/я/c', expected='../b'),
            dict(url='a/я/b', other='a/яя/c', expected='../../я/b'),
        ]:
            url, other, expected = case['url'], case['other'], case['expected']
            with self.subTest(url=url, other=other):
                # Leading slash intentionally ignored
                self.assertEqual(utils.get_relative_url(url, other), expected)
                self.assertEqual(utils.get_relative_url('/' + url, other), expected)
                self.assertEqual(utils.get_relative_url(url, '/' + other), expected)
                self.assertEqual(utils.get_relative_url('/' + url, '/' + other), expected)

    def test_get_relative_url_empty(self):
        for url in ['', '.', '/.']:
            for other in ['', '.', '/', '/.']:
                with self.subTest(url=url, other=other):
                    self.assertEqual(utils.get_relative_url(url, other), '.')

        self.assertEqual(utils.get_relative_url('/', ''), './')
        self.assertEqual(utils.get_relative_url('/', '/'), './')
        self.assertEqual(utils.get_relative_url('/', '.'), './')
        self.assertEqual(utils.get_relative_url('/', '/.'), './')

    def test_normalize_url(self):
        def test(path, base, expected):
            self.assertEqual(utils.normalize_url(path, _Page(base)), expected)

        # Absolute paths and anchors are unaffected.
        for base in 'about.html', 'foo/bar.html', 'index.html', '', 'about/', 'foo/bar/':
            test('https://media.cdn.org/jq.js', base, 'https://media.cdn.org/jq.js')
            test('http://media.cdn.org/jquery.js', base, 'http://media.cdn.org/jquery.js')
            test('//media.cdn.org/jquery.js', base, '//media.cdn.org/jquery.js')
            test('#some_id', base, '#some_id')

        path = 'media.cdn.org/jquery.js'
        test(path, '', 'media.cdn.org/jquery.js')
        test(path, 'index.html', 'media.cdn.org/jquery.js')
        test(path, 'about.html', 'media.cdn.org/jquery.js')
        test(path, 'about/', '../media.cdn.org/jquery.js')
        test(path, 'foo/bar.html', '../media.cdn.org/jquery.js')
        test(path, 'foo/bar/', '../../media.cdn.org/jquery.js')

        path = 'local/file/jquery.js'
        test(path, '', 'local/file/jquery.js')
        test(path, 'index.html', 'local/file/jquery.js')
        test(path, 'about.html', 'local/file/jquery.js')
        test(path, 'about/', '../local/file/jquery.js')
        test(path, 'foo/bar.html', '../local/file/jquery.js')
        test(path, 'foo/bar/', '../../local/file/jquery.js')

        path = '../../../../above/jquery.js'
        test(path, '', '../../../../above/jquery.js')
        test(path, 'index.html', '../../../../above/jquery.js')
        test(path, 'about.html', '../../../../above/jquery.js')
        test(path, 'about/', '../../../../../above/jquery.js')
        test(path, 'foo/bar.html', '../../../../../above/jquery.js')
        test(path, 'foo/bar/', '../../../../../../above/jquery.js')

        path = '../some/dir/'
        test(path, '', '../some/dir/')
        test(path, 'index.html', '../some/dir/')
        test(path, 'about.html', '../some/dir/')
        test(path, 'about/', '../../some/dir/')
        test(path, 'foo/bar.html', '../../some/dir/')
        test(path, 'foo/bar/', '../../../some/dir/')

        path = 'image.png'
        test(path, '', 'image.png')
        test(path, 'index.html', 'image.png')
        test(path, 'about.html', 'image.png')
        test(path, 'about/', '../image.png')
        test(path, 'foo/bar.html', '../image.png')
        test(path, 'foo/bar/', '../../image.png')

        path = 'style.css?v=20180308c'
        test(path, '', 'style.css?v=20180308c')
        test(path, 'index.html', 'style.css?v=20180308c')
        test(path, 'about.html', 'style.css?v=20180308c')
        test(path, 'about/', '../style.css?v=20180308c')
        test(path, 'foo/bar.html', '../style.css?v=20180308c')
        test(path, 'foo/bar/', '../../style.css?v=20180308c')

    # TODO: This shouldn't pass on Linux
    # @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_normalize_url_windows(self):
        def test(path, base, expected):
            self.assertEqual(utils.normalize_url(path, _Page(base)), expected)

        with self.assertLogs('mkdocs', level='WARNING'):
            path = 'local\\windows\\file\\jquery.js'
            test(path, '', 'local/windows/file/jquery.js')
            test(path, 'about/', '../local/windows/file/jquery.js')
            test(path, 'foo/bar/', '../../local/windows/file/jquery.js')

    def test_reduce_list(self):
        self.assertEqual(
            utils.reduce_list([1, 2, 3, 4, 5, 5, 2, 4, 6, 7, 8]),
            [1, 2, 3, 4, 5, 6, 7, 8],
        )

    def test_insort(self):
        a = [1, 2, 3]
        utils.insort(a, 5)
        self.assertEqual(a, [1, 2, 3, 5])
        utils.insort(a, -1)
        self.assertEqual(a, [-1, 1, 2, 3, 5])
        utils.insort(a, 2)
        self.assertEqual(a, [-1, 1, 2, 2, 3, 5])
        utils.insort(a, 4)
        self.assertEqual(a, [-1, 1, 2, 2, 3, 4, 5])

    def test_insort_key(self):
        a = [(1, 'a'), (1, 'b'), (2, 'c')]
        utils.insort(a, (1, 'a'), key=lambda v: v[0])
        self.assertEqual(a, [(1, 'a'), (1, 'b'), (1, 'a'), (2, 'c')])

    def test_nest_paths(self, j=posixpath.join):
        result = utils.nest_paths(
            [
                'index.md',
                j('user-guide', 'configuration.md'),
                j('user-guide', 'styling-your-docs.md'),
                j('user-guide', 'writing-your-docs.md'),
                j('about', 'contributing.md'),
                j('about', 'license.md'),
                j('about', 'release-notes.md'),
            ]
        )

        self.assertEqual(
            result,
            [
                'index.md',
                {
                    'User guide': [
                        j('user-guide', 'configuration.md'),
                        j('user-guide', 'styling-your-docs.md'),
                        j('user-guide', 'writing-your-docs.md'),
                    ]
                },
                {
                    'About': [
                        j('about', 'contributing.md'),
                        j('about', 'license.md'),
                        j('about', 'release-notes.md'),
                    ]
                },
            ],
        )

    def test_nest_paths_native(self):
        self.test_nest_paths(os.path.join)

    def test_unicode_yaml(self):
        yaml_src = dedent(
            '''
            key: value
            key2:
              - value
            '''
        )

        config = utils.yaml_load(yaml_src)
        self.assertTrue(isinstance(config['key'], str))
        self.assertTrue(isinstance(config['key2'][0], str))

    @mock.patch.dict(os.environ, {'VARNAME': 'Hello, World!', 'BOOLVAR': 'false'})
    def test_env_var_in_yaml(self):
        yaml_src = dedent(
            '''
            key1: !ENV VARNAME
            key2: !ENV UNDEFINED
            key3: !ENV [UNDEFINED, default]
            key4: !ENV [UNDEFINED, VARNAME, default]
            key5: !ENV BOOLVAR
            '''
        )
        config = utils.yaml_load(yaml_src)
        self.assertEqual(config['key1'], 'Hello, World!')
        self.assertIsNone(config['key2'])
        self.assertEqual(config['key3'], 'default')
        self.assertEqual(config['key4'], 'Hello, World!')
        self.assertIs(config['key5'], False)

    @tempdir(files={'base.yml': BASEYML, 'parent.yml': PARENTYML})
    def test_yaml_inheritance(self, tdir):
        expected = {
            'foo': 'bar',
            'baz': {
                'sub1': 'replaced',
                'sub2': 2,
                'sub3': 'new',
            },
            'deep1': {
                'deep2-1': {
                    'deep3-1': 'replaced',
                    'deep3-2': 'bar',
                },
                'deep2-2': 'baz',
            },
        }
        with open(os.path.join(tdir, 'base.yml')) as fd:
            result = utils.yaml_load(fd)
        self.assertEqual(result, expected)

    @tempdir(files={'base.yml': BASEYML})
    def test_yaml_inheritance_missing_parent(self, tdir):
        with open(os.path.join(tdir, 'base.yml')) as fd:
            with self.assertRaises(exceptions.ConfigurationError):
                utils.yaml_load(fd)

    @tempdir()
    @tempdir()
    def test_copy_files(self, src_dir, dst_dir):
        cases = [
            dict(
                src_path='foo.txt',
                dst_path='foo.txt',
                expected='foo.txt',
            ),
            dict(
                src_path='bar.txt',
                dst_path='foo/',  # ensure src filename is appended
                expected='foo/bar.txt',
            ),
            dict(
                src_path='baz.txt',
                dst_path='foo/bar/baz.txt',  # ensure missing dirs are created
                expected='foo/bar/baz.txt',
            ),
        ]

        for case in cases:
            src, dst, expected = case['src_path'], case['dst_path'], case['expected']
            with self.subTest(src):
                src = os.path.join(src_dir, src)
                with open(src, 'w') as f:
                    f.write('content')
                dst = os.path.join(dst_dir, dst)
                utils.copy_file(src, dst)
                self.assertTrue(os.path.isfile(os.path.join(dst_dir, expected)))

    @tempdir()
    @tempdir()
    def test_copy_files_without_permissions(self, src_dir, dst_dir):
        cases = [
            dict(src_path='foo.txt', expected='foo.txt'),
            dict(src_path='bar.txt', expected='bar.txt'),
            dict(src_path='baz.txt', expected='baz.txt'),
        ]

        try:
            for case in cases:
                src, expected = case['src_path'], case['expected']
                with self.subTest(src):
                    src = os.path.join(src_dir, src)
                    with open(src, 'w') as f:
                        f.write('content')
                    # Set src file to read-only
                    os.chmod(src, stat.S_IRUSR)
                    utils.copy_file(src, dst_dir)
                    self.assertTrue(os.path.isfile(os.path.join(dst_dir, expected)))
                    self.assertNotEqual(
                        os.stat(src).st_mode, os.stat(os.path.join(dst_dir, expected)).st_mode
                    )
                    # While src was read-only, dst must remain writable
                    self.assertTrue(os.access(os.path.join(dst_dir, expected), os.W_OK))
        finally:
            for case in cases:
                # Undo read-only so we can delete temp files
                src = os.path.join(src_dir, case['src_path'])
                if os.path.exists(src):
                    os.chmod(src, stat.S_IRUSR | stat.S_IWUSR)

    def test_mm_meta_data(self):
        doc = dedent(
            """
            Title: Foo Bar
            Date: 2018-07-10
            Summary: Line one
                Line two
            Tags: foo
            Tags: bar

            Doc body
            """
        )
        self.assertEqual(
            meta.get_data(doc),
            (
                "Doc body",
                {
                    'title': 'Foo Bar',
                    'date': '2018-07-10',
                    'summary': 'Line one Line two',
                    'tags': 'foo bar',
                },
            ),
        )

    def test_mm_meta_data_blank_first_line(self):
        doc = '\nfoo: bar\nDoc body'
        self.assertEqual(meta.get_data(doc), (doc.lstrip(), {}))

    def test_yaml_meta_data(self):
        doc = dedent(
            """
            ---
            Title: Foo Bar
            Date: 2018-07-10
            Summary: Line one
                Line two
            Tags:
                - foo
                - bar
            ---
            Doc body
            """
        )
        self.assertEqual(
            meta.get_data(doc),
            (
                "Doc body",
                {
                    'Title': 'Foo Bar',
                    'Date': datetime.date(2018, 7, 10),
                    'Summary': 'Line one Line two',
                    'Tags': ['foo', 'bar'],
                },
            ),
        )

    def test_yaml_meta_data_not_dict(self):
        doc = dedent(
            """
            ---
            - List item
            ---
            Doc body
            """
        )
        self.assertEqual(meta.get_data(doc), (doc, {}))

    def test_yaml_meta_data_invalid(self):
        doc = dedent(
            """
            ---
            foo: bar: baz
            ---
            Doc body
            """
        )
        self.assertEqual(meta.get_data(doc), (doc, {}))

    def test_no_meta_data(self):
        doc = dedent(
            """
            Doc body
            """
        )
        self.assertEqual(meta.get_data(doc), (doc, {}))


class ThemeUtilsTests(unittest.TestCase):
    def setUp(self):
        utils.get_themes.cache_clear()

    def test_get_themes(self):
        themes = utils.get_theme_names()
        self.assertIn('mkdocs', themes)
        self.assertIn('readthedocs', themes)

    @mock.patch('mkdocs.utils.entry_points', autospec=True)
    def test_get_theme_dir(self, mock_iter):
        path = 'some/path'

        theme = mock.Mock()
        theme.name = 'mkdocs2'
        theme.dist.name = 'mkdocs2'
        theme.load().__file__ = os.path.join(path, '__init__.py')

        mock_iter.return_value = [theme]

        self.assertEqual(utils.get_theme_dir(theme.name), os.path.abspath(path))

    def test_get_theme_dir_error(self):
        with self.assertRaises(KeyError):
            utils.get_theme_dir('nonexistanttheme')

    @mock.patch('mkdocs.utils.entry_points', autospec=True)
    def test_get_theme_dir_importerror(self, mock_iter):
        theme = mock.Mock()
        theme.name = 'mkdocs2'
        theme.dist.name = 'mkdocs2'
        theme.load.side_effect = ImportError()

        mock_iter.return_value = [theme]

        with self.assertRaises(ImportError):
            utils.get_theme_dir(theme.name)

    @mock.patch('mkdocs.utils.entry_points', autospec=True)
    def test_get_themes_warning(self, mock_iter):
        theme1 = mock.Mock()
        theme1.name = 'mkdocs2'
        theme1.dist.name = 'mkdocs2'
        theme1.load().__file__ = "some/path1"

        theme2 = mock.Mock()
        theme2.name = 'mkdocs2'
        theme2.dist.name = 'mkdocs3'
        theme2.load().__file__ = "some/path2"

        mock_iter.return_value = [theme1, theme2]

        with self.assertLogs('mkdocs') as cm:
            theme_names = utils.get_theme_names()
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.utils:A theme named 'mkdocs2' is provided by the Python "
            "packages 'mkdocs3' and 'mkdocs2'. The one in 'mkdocs3' will be used.",
        )
        self.assertCountEqual(theme_names, ['mkdocs2'])

    @mock.patch('mkdocs.utils.entry_points', autospec=True)
    def test_get_themes_error(self, mock_iter):
        theme1 = mock.Mock()
        theme1.name = 'mkdocs'
        theme1.dist.name = 'mkdocs'
        theme1.load().__file__ = "some/path1"

        theme2 = mock.Mock()
        theme2.name = 'mkdocs'
        theme2.dist.name = 'mkdocs2'
        theme2.load().__file__ = "some/path2"

        mock_iter.return_value = [theme1, theme2]

        with self.assertRaisesRegex(
            exceptions.ConfigurationError,
            "The theme 'mkdocs' is a builtin theme but the package 'mkdocs2' "
            "attempts to provide a theme with the same name.",
        ):
            utils.get_theme_names()


class LogCounterTests(unittest.TestCase):
    def setUp(self):
        self.log = logging.getLogger('dummy')
        self.log.propagate = False
        self.log.setLevel(1)
        self.counter = utils.CountHandler()
        self.log.addHandler(self.counter)

    def tearDown(self):
        self.log.removeHandler(self.counter)

    def test_default_values(self):
        self.assertEqual(self.counter.get_counts(), [])

    def test_count_critical(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.critical('msg')
        self.assertEqual(self.counter.get_counts(), [('CRITICAL', 1)])

    def test_count_error(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.error('msg')
        self.assertEqual(self.counter.get_counts(), [('ERROR', 1)])

    def test_count_warning(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.warning('msg')
        self.assertEqual(self.counter.get_counts(), [('WARNING', 1)])

    def test_count_info(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.info('msg')
        self.assertEqual(self.counter.get_counts(), [('INFO', 1)])

    def test_count_debug(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.debug('msg')
        self.assertEqual(self.counter.get_counts(), [('DEBUG', 1)])

    def test_count_multiple(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.log.warning('msg 1')
        self.assertEqual(self.counter.get_counts(), [('WARNING', 1)])
        self.log.warning('msg 2')
        self.assertEqual(self.counter.get_counts(), [('WARNING', 2)])
        self.log.debug('msg 3')
        self.assertEqual(self.counter.get_counts(), [('WARNING', 2), ('DEBUG', 1)])
        self.log.error('mdg 4')
        self.assertEqual(self.counter.get_counts(), [('ERROR', 1), ('WARNING', 2), ('DEBUG', 1)])

    def test_log_level(self):
        self.assertEqual(self.counter.get_counts(), [])
        self.counter.setLevel(logging.ERROR)
        self.log.error('counted')
        self.log.warning('not counted')
        self.log.info('not counted')
        self.assertEqual(self.counter.get_counts(), [('ERROR', 1)])
        self.counter.setLevel(logging.WARNING)
        self.log.error('counted')
        self.log.warning('counted')
        self.log.info('not counted')
        self.assertEqual(self.counter.get_counts(), [('ERROR', 2), ('WARNING', 1)])


@dataclasses.dataclass
class _Page:
    url: str



import unittest
from unittest import mock

from ghp_import import GhpError  # type: ignore

from mkdocs import __version__
from mkdocs.commands import gh_deploy
from mkdocs.exceptions import Abort
from mkdocs.tests.base import load_config


class TestGitHubDeploy(unittest.TestCase):
    @mock.patch('subprocess.Popen')
    def test_is_cwd_git_repo(self, mock_popeno):
        mock_popeno().wait.return_value = 0

        self.assertTrue(gh_deploy._is_cwd_git_repo())

    @mock.patch('subprocess.Popen')
    def test_is_cwd_not_git_repo(self, mock_popeno):
        mock_popeno().wait.return_value = 1

        self.assertFalse(gh_deploy._is_cwd_git_repo())

    @mock.patch('subprocess.Popen')
    def test_get_current_sha(self, mock_popeno):
        mock_popeno().communicate.return_value = (b'6d98394\n', b'')

        self.assertEqual(gh_deploy._get_current_sha('.'), '6d98394')

    @mock.patch('subprocess.Popen')
    def test_get_remote_url_ssh(self, mock_popeno):
        mock_popeno().communicate.return_value = (
            b'git@github.com:mkdocs/mkdocs.git\n',
            b'',
        )

        expected = ('git@', 'mkdocs/mkdocs.git')
        self.assertEqual(expected, gh_deploy._get_remote_url('origin'))

    @mock.patch('subprocess.Popen')
    def test_get_remote_url_http(self, mock_popeno):
        mock_popeno().communicate.return_value = (
            b'https://github.com/mkdocs/mkdocs.git\n',
            b'',
        )

        expected = ('https://', 'mkdocs/mkdocs.git')
        self.assertEqual(expected, gh_deploy._get_remote_url('origin'))

    @mock.patch('subprocess.Popen')
    def test_get_remote_url_enterprise(self, mock_popeno):
        mock_popeno().communicate.return_value = (
            b'https://notgh.com/mkdocs/mkdocs.git\n',
            b'',
        )

        expected = (None, None)
        self.assertEqual(expected, gh_deploy._get_remote_url('origin'))

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch('mkdocs.commands.gh_deploy._get_remote_url', mock.Mock(return_value=(None, None)))
    @mock.patch('mkdocs.commands.gh_deploy._check_version', mock.Mock())
    @mock.patch('ghp_import.ghp_import', mock.Mock())
    def test_deploy(self):
        config = load_config(
            remote_branch='test',
        )
        gh_deploy.gh_deploy(config)

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch('mkdocs.commands.gh_deploy._get_remote_url', mock.Mock(return_value=(None, None)))
    @mock.patch('mkdocs.commands.gh_deploy._check_version', mock.Mock())
    @mock.patch('ghp_import.ghp_import', mock.Mock())
    @mock.patch('os.path.isfile', mock.Mock(return_value=False))
    def test_deploy_no_cname(self):
        config = load_config(
            remote_branch='test',
        )
        gh_deploy.gh_deploy(config)

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch(
        'mkdocs.commands.gh_deploy._get_remote_url',
        mock.Mock(return_value=('git@', 'mkdocs/mkdocs.git')),
    )
    @mock.patch('mkdocs.commands.gh_deploy._check_version', mock.Mock())
    @mock.patch('ghp_import.ghp_import', mock.Mock())
    def test_deploy_hostname(self):
        config = load_config(
            remote_branch='test',
        )
        gh_deploy.gh_deploy(config)

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch('mkdocs.commands.gh_deploy._get_remote_url', mock.Mock(return_value=(None, None)))
    @mock.patch('mkdocs.commands.gh_deploy._check_version')
    @mock.patch('ghp_import.ghp_import', mock.Mock())
    def test_deploy_ignore_version_default(self, check_version):
        config = load_config(
            remote_branch='test',
        )
        gh_deploy.gh_deploy(config)
        check_version.assert_called_once()

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch('mkdocs.commands.gh_deploy._get_remote_url', mock.Mock(return_value=(None, None)))
    @mock.patch('mkdocs.commands.gh_deploy._check_version')
    @mock.patch('ghp_import.ghp_import', mock.Mock())
    def test_deploy_ignore_version(self, check_version):
        config = load_config(
            remote_branch='test',
        )
        gh_deploy.gh_deploy(config, ignore_version=True)
        check_version.assert_not_called()

    @mock.patch('mkdocs.commands.gh_deploy._is_cwd_git_repo', mock.Mock(return_value=True))
    @mock.patch('mkdocs.commands.gh_deploy._get_current_sha', mock.Mock(return_value='shashas'))
    @mock.patch('mkdocs.commands.gh_deploy._check_version', mock.Mock())
    @mock.patch('ghp_import.ghp_import', mock.Mock(side_effect=GhpError('TestError123')))
    def test_deploy_error(self):
        config = load_config(
            remote_branch='test',
        )

        with self.assertLogs('mkdocs', level='ERROR') as cm:
            with self.assertRaises(Abort):
                gh_deploy.gh_deploy(config)
        self.assertEqual(
            cm.output,
            [
                'ERROR:mkdocs.commands.gh_deploy:Failed to deploy to GitHub with error: \n'
                'TestError123'
            ],
        )


class TestGitHubDeployLogs(unittest.TestCase):
    @mock.patch('subprocess.Popen')
    def test_mkdocs_newer(self, mock_popeno):
        mock_popeno().communicate.return_value = (
            b'Deployed 12345678 with MkDocs version: 0.1.2\n',
            b'',
        )

        with self.assertLogs('mkdocs') as cm:
            gh_deploy._check_version('gh-pages')
        self.assertEqual(
            '\n'.join(cm.output),
            f'INFO:mkdocs.commands.gh_deploy:Previous deployment was done with MkDocs '
            f'version 0.1.2; you are deploying with a newer version ({__version__})',
        )

    @mock.patch('subprocess.Popen')
    def test_mkdocs_older(self, mock_popeno):
        mock_popeno().communicate.return_value = (
            b'Deployed 12345678 with MkDocs version: 10.1.2\n',
            b'',
        )

        with self.assertLogs('mkdocs', level='ERROR') as cm:
            with self.assertRaises(Abort):
                gh_deploy._check_version('gh-pages')
        self.assertEqual(
            '\n'.join(cm.output),
            f'ERROR:mkdocs.commands.gh_deploy:Deployment terminated: Previous deployment was made with '
            f'MkDocs version 10.1.2; you are attempting to deploy with an older version ({__version__}).'
            f' Use --ignore-version to deploy anyway.',
        )

    @mock.patch('subprocess.Popen')
    def test_version_unknown(self, mock_popeno):
        mock_popeno().communicate.return_value = (b'No version specified\n', b'')

        with self.assertLogs('mkdocs') as cm:
            gh_deploy._check_version('gh-pages')
        self.assertEqual(
            '\n'.join(cm.output),
            'WARNING:mkdocs.commands.gh_deploy:Version check skipped: No version specified in previous deployment.',
        )



from __future__ import annotations

import enum
import fnmatch
import logging
import os
import posixpath
import shutil
import warnings
from pathlib import PurePath
from typing import TYPE_CHECKING, Callable, Iterable, Iterator, Sequence
from urllib.parse import quote as urlquote

import pathspec
import pathspec.gitignore
import pathspec.util

from mkdocs import utils

if TYPE_CHECKING:
    import jinja2.environment

    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.pages import Page


log = logging.getLogger(__name__)


class InclusionLevel(enum.Enum):
    EXCLUDED = -3
    """The file is excluded and will not be processed."""
    DRAFT = -2
    """The file is excluded from the final site, but will still be populated during `mkdocs serve`."""
    NOT_IN_NAV = -1
    """The file is part of the site, but doesn't produce nav warnings."""
    UNDEFINED = 0
    """Still needs to be computed based on the config. If the config doesn't kick in, acts the same as `included`."""
    INCLUDED = 1
    """The file is part of the site. Documentation pages that are omitted from the nav will produce warnings."""

    def all(self):
        return True

    def is_included(self):
        return self.value > self.DRAFT.value

    def is_excluded(self):
        return self.value <= self.DRAFT.value

    def is_in_serve(self):
        return self.value >= self.DRAFT.value

    def is_in_nav(self):
        return self.value > self.NOT_IN_NAV.value

    def is_not_in_nav(self):
        return self.value <= self.NOT_IN_NAV.value


class Files:
    """A collection of [File][mkdocs.structure.files.File] objects."""

    def __init__(self, files: list[File]) -> None:
        self._files = files
        self._src_uris: dict[str, File] | None = None

    def __iter__(self) -> Iterator[File]:
        """Iterate over the files within."""
        return iter(self._files)

    def __len__(self) -> int:
        """The number of files within."""
        return len(self._files)

    def __contains__(self, path: str) -> bool:
        """Whether the file with this `src_uri` is in the collection."""
        return PurePath(path).as_posix() in self.src_uris

    @property
    def src_paths(self) -> dict[str, File]:
        """Soft-deprecated, prefer `src_uris`."""
        return {file.src_path: file for file in self._files}

    @property
    def src_uris(self) -> dict[str, File]:
        """
        A mapping containing every file, with the keys being their
        [`src_uri`][mkdocs.structure.files.File.src_uri].
        """
        if self._src_uris is None:
            self._src_uris = {file.src_uri: file for file in self._files}
        return self._src_uris

    def get_file_from_path(self, path: str) -> File | None:
        """Return a File instance with File.src_uri equal to path."""
        return self.src_uris.get(PurePath(path).as_posix())

    def append(self, file: File) -> None:
        """Append file to Files collection."""
        self._src_uris = None
        self._files.append(file)

    def remove(self, file: File) -> None:
        """Remove file from Files collection."""
        self._src_uris = None
        self._files.remove(file)

    def copy_static_files(
        self,
        dirty: bool = False,
        *,
        inclusion: Callable[[InclusionLevel], bool] = InclusionLevel.is_included,
    ) -> None:
        """Copy static files from source to destination."""
        for file in self:
            if not file.is_documentation_page() and inclusion(file.inclusion):
                file.copy_file(dirty)

    def documentation_pages(
        self, *, inclusion: Callable[[InclusionLevel], bool] = InclusionLevel.is_included
    ) -> Sequence[File]:
        """Return iterable of all Markdown page file objects."""
        return [file for file in self if file.is_documentation_page() and inclusion(file.inclusion)]

    def static_pages(self) -> Sequence[File]:
        """Return iterable of all static page file objects."""
        return [file for file in self if file.is_static_page()]

    def media_files(self) -> Sequence[File]:
        """Return iterable of all file objects which are not documentation or static pages."""
        return [file for file in self if file.is_media_file()]

    def javascript_files(self) -> Sequence[File]:
        """Return iterable of all javascript file objects."""
        return [file for file in self if file.is_javascript()]

    def css_files(self) -> Sequence[File]:
        """Return iterable of all CSS file objects."""
        return [file for file in self if file.is_css()]

    def add_files_from_theme(self, env: jinja2.Environment, config: MkDocsConfig) -> None:
        """Retrieve static files from Jinja environment and add to collection."""

        def filter(name):
            # '.*' filters dot files/dirs at root level whereas '*/.*' filters nested levels
            patterns = ['.*', '*/.*', '*.py', '*.pyc', '*.html', '*readme*', 'mkdocs_theme.yml']
            # Exclude translation files
            patterns.append("locales/*")
            patterns.extend(f'*{x}' for x in utils.markdown_extensions)
            patterns.extend(config.theme.static_templates)
            for pattern in patterns:
                if fnmatch.fnmatch(name.lower(), pattern):
                    return False
            return True

        for path in env.list_templates(filter_func=filter):
            # Theme files do not override docs_dir files
            path = PurePath(path).as_posix()
            if path not in self.src_uris:
                for dir in config.theme.dirs:
                    # Find the first theme dir which contains path
                    if os.path.isfile(os.path.join(dir, path)):
                        self.append(File(path, dir, config.site_dir, config.use_directory_urls))
                        break


class File:
    """
    A MkDocs File object.

    Points to the source and destination locations of a file.

    The `path` argument must be a path that exists relative to `src_dir`.

    The `src_dir` and `dest_dir` must be absolute paths on the local file system.

    The `use_directory_urls` argument controls how destination paths are generated. If `False`, a Markdown file is
    mapped to an HTML file of the same name (the file extension is changed to `.html`). If True, a Markdown file is
    mapped to an HTML index file (`index.html`) nested in a directory using the "name" of the file in `path`. The
    `use_directory_urls` argument has no effect on non-Markdown files.

    File objects have the following properties, which are Unicode strings:
    """

    src_uri: str
    """The pure path (always '/'-separated) of the source file relative to the source directory."""

    abs_src_path: str
    """The absolute concrete path of the source file. Will use backslashes on Windows."""

    dest_uri: str
    """The pure path (always '/'-separated) of the destination file relative to the destination directory."""

    abs_dest_path: str
    """The absolute concrete path of the destination file. Will use backslashes on Windows."""

    url: str
    """The URI of the destination file relative to the destination directory as a string."""

    inclusion: InclusionLevel = InclusionLevel.UNDEFINED
    """Whether the file will be excluded from the built site."""

    @property
    def src_path(self) -> str:
        """Same as `src_uri` (and synchronized with it) but will use backslashes on Windows. Discouraged."""
        return os.path.normpath(self.src_uri)

    @src_path.setter
    def src_path(self, value):
        self.src_uri = PurePath(value).as_posix()

    @property
    def dest_path(self) -> str:
        """Same as `dest_uri` (and synchronized with it) but will use backslashes on Windows. Discouraged."""
        return os.path.normpath(self.dest_uri)

    @dest_path.setter
    def dest_path(self, value):
        self.dest_uri = PurePath(value).as_posix()

    page: Page | None

    def __init__(
        self,
        path: str,
        src_dir: str,
        dest_dir: str,
        use_directory_urls: bool,
        *,
        dest_uri: str | None = None,
        inclusion: InclusionLevel = InclusionLevel.UNDEFINED,
    ) -> None:
        self.page = None
        self.src_path = path
        self.name = self._get_stem()
        if dest_uri is None:
            dest_uri = self._get_dest_path(use_directory_urls)
        self.dest_uri = dest_uri
        self.url = self._get_url(use_directory_urls)
        self.abs_src_path = os.path.normpath(os.path.join(src_dir, self.src_uri))
        self.abs_dest_path = os.path.normpath(os.path.join(dest_dir, self.dest_uri))
        self.inclusion = inclusion

    def __repr__(self):
        return (
            f"File(src_uri='{self.src_uri}', dest_uri='{self.dest_uri}',"
            f" name='{self.name}', url='{self.url}')"
        )

    def _get_stem(self) -> str:
        """Return the name of the file without its extension."""
        filename = posixpath.basename(self.src_uri)
        stem, ext = posixpath.splitext(filename)
        return 'index' if stem == 'README' else stem

    def _get_dest_path(self, use_directory_urls: bool) -> str:
        """Return destination path based on source path."""
        if self.is_documentation_page():
            parent, filename = posixpath.split(self.src_uri)
            if not use_directory_urls or self.name == 'index':
                # index.md or README.md => index.html
                # foo.md => foo.html
                return posixpath.join(parent, self.name + '.html')
            else:
                # foo.md => foo/index.html
                return posixpath.join(parent, self.name, 'index.html')
        return self.src_uri

    def _get_url(self, use_directory_urls: bool) -> str:
        """Return url based in destination path."""
        url = self.dest_uri
        dirname, filename = posixpath.split(url)
        if use_directory_urls and filename == 'index.html':
            url = (dirname or '.') + '/'
        return urlquote(url)

    def url_relative_to(self, other: File | str) -> str:
        """Return url for file relative to other file."""
        return utils.get_relative_url(self.url, other.url if isinstance(other, File) else other)

    def copy_file(self, dirty: bool = False) -> None:
        """Copy source file to destination, ensuring parent directories exist."""
        if dirty and not self.is_modified():
            log.debug(f"Skip copying unmodified file: '{self.src_uri}'")
        else:
            log.debug(f"Copying media file: '{self.src_uri}'")
            try:
                utils.copy_file(self.abs_src_path, self.abs_dest_path)
            except shutil.SameFileError:
                pass  # Let plugins write directly into site_dir.

    def is_modified(self) -> bool:
        if os.path.isfile(self.abs_dest_path):
            return os.path.getmtime(self.abs_dest_path) < os.path.getmtime(self.abs_src_path)
        return True

    def is_documentation_page(self) -> bool:
        """Return True if file is a Markdown page."""
        return utils.is_markdown_file(self.src_uri)

    def is_static_page(self) -> bool:
        """Return True if file is a static page (HTML, XML, JSON)."""
        return self.src_uri.endswith(('.html', '.htm', '.xml', '.json'))

    def is_media_file(self) -> bool:
        """Return True if file is not a documentation or static page."""
        return not (self.is_documentation_page() or self.is_static_page())

    def is_javascript(self) -> bool:
        """Return True if file is a JavaScript file."""
        return self.src_uri.endswith(('.js', '.javascript', '.mjs'))

    def is_css(self) -> bool:
        """Return True if file is a CSS file."""
        return self.src_uri.endswith('.css')


_default_exclude = pathspec.gitignore.GitIgnoreSpec.from_lines(['.*', '/templates/'])


def set_exclusions(files: Iterable[File], config: MkDocsConfig) -> None:
    """Re-calculate which files are excluded, based on the patterns in the config."""
    exclude: pathspec.gitignore.GitIgnoreSpec | None = config.get('exclude_docs')
    exclude = _default_exclude + exclude if exclude else _default_exclude
    drafts: pathspec.gitignore.GitIgnoreSpec | None = config.get('draft_docs')
    nav_exclude: pathspec.gitignore.GitIgnoreSpec | None = config.get('not_in_nav')

    for file in files:
        if file.inclusion == InclusionLevel.UNDEFINED:
            if exclude.match_file(file.src_uri):
                file.inclusion = InclusionLevel.EXCLUDED
            elif drafts and drafts.match_file(file.src_uri):
                file.inclusion = InclusionLevel.DRAFT
            elif nav_exclude and nav_exclude.match_file(file.src_uri):
                file.inclusion = InclusionLevel.NOT_IN_NAV
            else:
                file.inclusion = InclusionLevel.INCLUDED


def get_files(config: MkDocsConfig) -> Files:
    """Walk the `docs_dir` and return a Files collection."""
    files: list[File] = []
    conflicting_files: list[tuple[File, File]] = []
    for source_dir, dirnames, filenames in os.walk(config['docs_dir'], followlinks=True):
        relative_dir = os.path.relpath(source_dir, config['docs_dir'])
        dirnames.sort()
        filenames.sort(key=_file_sort_key)

        files_by_dest: dict[str, File] = {}
        for filename in filenames:
            file = File(
                os.path.join(relative_dir, filename),
                config['docs_dir'],
                config['site_dir'],
                config['use_directory_urls'],
            )
            # Skip README.md if an index file also exists in dir (part 1)
            prev_file = files_by_dest.setdefault(file.dest_uri, file)
            if prev_file is not file:
                conflicting_files.append((prev_file, file))
            files.append(file)
            prev_file = file

    set_exclusions(files, config)
    # Skip README.md if an index file also exists in dir (part 2)
    for a, b in conflicting_files:
        if b.inclusion.is_included():
            if a.inclusion.is_included():
                log.warning(
                    f"Excluding '{a.src_uri}' from the site because it conflicts with '{b.src_uri}'."
                )
            try:
                files.remove(a)
            except ValueError:
                pass  # Catching this to avoid errors if attempting to remove the same file twice.
        else:
            try:
                files.remove(b)
            except ValueError:
                pass

    return Files(files)


def _file_sort_key(f: str):
    """Always sort `index` or `README` as first filename in list."""
    return (os.path.splitext(f)[0] not in ('index', 'README'), f)


def _sort_files(filenames: Iterable[str]) -> list[str]:
    return sorted(filenames, key=_file_sort_key)


def _filter_paths(basename: str, path: str, is_dir: bool, exclude: Iterable[str]) -> bool:
    warnings.warn(
        "_filter_paths is not used since MkDocs 1.5 and will be removed soon.", DeprecationWarning
    )
    for item in exclude:
        # Items ending in '/' apply only to directories.
        if item.endswith('/') and not is_dir:
            continue
        # Items starting with '/' apply to the whole path.
        # In any other cases just the basename is used.
        match = path if item.startswith('/') else basename
        if fnmatch.fnmatch(match, item.strip('/')):
            return True
    return False



from __future__ import annotations

import logging
import os
import warnings
from typing import Any, Collection, MutableMapping

import jinja2
import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader  # type: ignore

from mkdocs import localization, utils
from mkdocs.config.base import ValidationError
from mkdocs.utils import templates

log = logging.getLogger(__name__)


class Theme(MutableMapping[str, Any]):
    """
    A Theme object.

    Args:
        name: The name of the theme as defined by its entrypoint.
        custom_dir: User defined directory for custom templates.
        static_templates: A list of templates to render as static pages.

    All other keywords are passed as-is and made available as a key/value mapping.
    """

    def __init__(
        self,
        name: str | None = None,
        *,
        custom_dir: str | None = None,
        static_templates: Collection[str] = (),
        locale: str | None = None,
        **user_config,
    ) -> None:
        self.name = name
        self._custom_dir = custom_dir
        _vars: dict[str, Any] = {'name': name, 'locale': 'en'}
        self.__vars = _vars

        # MkDocs provided static templates are always included
        package_dir = os.path.abspath(os.path.dirname(__file__))
        mkdocs_templates = os.path.join(package_dir, 'templates')
        self.static_templates = set(os.listdir(mkdocs_templates))

        # Build self.dirs from various sources in order of precedence
        self.dirs = []

        if custom_dir is not None:
            self.dirs.append(custom_dir)

        if name:
            self._load_theme_config(name)

        # Include templates provided directly by MkDocs (outside any theme)
        self.dirs.append(mkdocs_templates)

        # Handle remaining user configs. Override theme configs (if set)
        self.static_templates.update(static_templates)
        _vars.update(user_config)

        # Validate locale and convert to Locale object
        if locale is None:
            locale = _vars['locale']
        _vars['locale'] = localization.parse_locale(locale)

    name: str | None

    @property
    def locale(self) -> localization.Locale:
        return self['locale']

    @property
    def custom_dir(self) -> str | None:
        return self._custom_dir

    @property
    def _vars(self) -> dict[str, Any]:
        warnings.warn(
            "Do not access Theme._vars, instead access the keys of Theme directly.",
            DeprecationWarning,
        )
        return self.__vars

    dirs: list[str]

    static_templates: set[str]

    def __repr__(self) -> str:
        return "{}(name={!r}, dirs={!r}, static_templates={!r}, {})".format(
            self.__class__.__name__,
            self.name,
            self.dirs,
            self.static_templates,
            ', '.join(f'{k}={v!r}' for k, v in self.items()),
        )

    def __getitem__(self, key: str) -> Any:
        return self.__vars[key]

    def __setitem__(self, key: str, value):
        self.__vars[key] = value

    def __delitem__(self, key: str):
        del self.__vars[key]

    def __contains__(self, item: object) -> bool:
        return item in self.__vars

    def __len__(self):
        return len(self.__vars)

    def __iter__(self):
        return iter(self.__vars)

    def _load_theme_config(self, name: str) -> None:
        """Recursively load theme and any parent themes."""
        theme_dir = utils.get_theme_dir(name)
        utils.get_themes.cache_clear()
        self.dirs.append(theme_dir)

        try:
            file_path = os.path.join(theme_dir, 'mkdocs_theme.yml')
            with open(file_path, 'rb') as f:
                theme_config = yaml.load(f, SafeLoader)
        except OSError as e:
            log.debug(e)
            raise ValidationError(
                f"The theme '{name}' does not appear to have a configuration file. "
                f"Please upgrade to a current version of the theme."
            )

        log.debug(f"Loaded theme configuration for '{name}' from '{file_path}': {theme_config}")

        if parent_theme := theme_config.pop('extends', None):
            themes = utils.get_theme_names()
            if parent_theme not in themes:
                raise ValidationError(
                    f"The theme '{name}' inherits from '{parent_theme}', which does not appear to be installed. "
                    f"The available installed themes are: {', '.join(themes)}"
                )
            self._load_theme_config(parent_theme)

        self.static_templates.update(theme_config.pop('static_templates', []))
        self.__vars.update(theme_config)

    def get_env(self) -> jinja2.Environment:
        """Return a Jinja environment for the theme."""
        loader = jinja2.FileSystemLoader(self.dirs)
        # No autoreload because editing a template in the middle of a build is not useful.
        env = jinja2.Environment(loader=loader, auto_reload=False)
        env.filters['url'] = templates.url_filter
        env.filters['script_tag'] = templates.script_tag_filter
        localization.install_translations(env, self.locale, self.dirs)
        return env



#!/usr/bin/env python

import os
import unittest

from mkdocs.commands import new
from mkdocs.tests.base import change_dir, tempdir


class NewTests(unittest.TestCase):
    @tempdir()
    def test_new(self, temp_dir):
        with change_dir(temp_dir):
            new.new("myproject")

            expected_paths = [
                os.path.join(temp_dir, "myproject"),
                os.path.join(temp_dir, "myproject", "mkdocs.yml"),
                os.path.join(temp_dir, "myproject", "docs"),
                os.path.join(temp_dir, "myproject", "docs", "index.md"),
            ]

            for expected_path in expected_paths:
                self.assertTrue(os.path.exists(expected_path))



import logging
import unittest.util

unittest.util._MAX_LENGTH = 100000


class DisallowLogsHandler(logging.Handler):
    def __init__(self, level=logging.WARNING):
        super().__init__(level=level)
        self.formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")

    def emit(self, record):
        raise AssertionError(f'Unexpected log: {self.format(record)!r}')


logging.lastResort = DisallowLogsHandler()  # type: ignore



import os
import sys
import unittest
from unittest import mock

from mkdocs.structure.files import File, Files, _sort_files, get_files
from mkdocs.tests.base import PathAssertionMixin, load_config, tempdir


class TestFiles(PathAssertionMixin, unittest.TestCase):
    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_src_path_windows(self):
        f = File('foo\\a.md', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(f.src_uri, 'foo/a.md')
        self.assertEqual(f.src_path, 'foo\\a.md')
        f.src_uri = 'foo/b.md'
        self.assertEqual(f.src_uri, 'foo/b.md')
        self.assertEqual(f.src_path, 'foo\\b.md')
        f.src_path = 'foo/c.md'
        self.assertEqual(f.src_uri, 'foo/c.md')
        self.assertEqual(f.src_path, 'foo\\c.md')
        f.src_path = 'foo\\d.md'
        self.assertEqual(f.src_uri, 'foo/d.md')
        self.assertEqual(f.src_path, 'foo\\d.md')
        f.src_uri = 'foo\\e.md'
        self.assertEqual(f.src_uri, 'foo\\e.md')
        self.assertEqual(f.src_path, 'foo\\e.md')

    def test_sort_files(self):
        self.assertEqual(
            _sort_files(['b.md', 'bb.md', 'a.md', 'index.md', 'aa.md']),
            ['index.md', 'a.md', 'aa.md', 'b.md', 'bb.md'],
        )

        self.assertEqual(
            _sort_files(['b.md', 'index.html', 'a.md', 'index.md']),
            ['index.html', 'index.md', 'a.md', 'b.md'],
        )

        self.assertEqual(
            _sort_files(['a.md', 'index.md', 'b.md', 'index.html']),
            ['index.html', 'index.md', 'a.md', 'b.md'],
        )

        self.assertEqual(
            _sort_files(['.md', '_.md', 'a.md', 'index.md', '1.md']),
            ['index.md', '.md', '1.md', '_.md', 'a.md'],
        )

        self.assertEqual(
            _sort_files(['a.md', 'b.md', 'a.md']),
            ['a.md', 'a.md', 'b.md'],
        )

        self.assertEqual(
            _sort_files(['A.md', 'B.md', 'README.md']),
            ['README.md', 'A.md', 'B.md'],
        )

    def test_md_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo.md', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo.md')
                if use_directory_urls:
                    self.assertEqual(f.dest_uri, 'foo/index.html')
                    self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/index.html')
                    self.assertEqual(f.url, 'foo/')
                else:
                    self.assertEqual(f.dest_uri, 'foo.html')
                    self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo.html')
                    self.assertEqual(f.url, 'foo.html')
                self.assertEqual(f.name, 'foo')
                self.assertTrue(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_md_file_nested(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/bar.md', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/bar.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/bar.md')
                if use_directory_urls:
                    self.assertEqual(f.dest_uri, 'foo/bar/index.html')
                    self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar/index.html')
                    self.assertEqual(f.url, 'foo/bar/')
                else:
                    self.assertEqual(f.dest_uri, 'foo/bar.html')
                    self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar.html')
                    self.assertEqual(f.url, 'foo/bar.html')
                self.assertEqual(f.name, 'bar')
                self.assertTrue(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_md_index_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('index.md', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'index.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/index.md')
                self.assertEqual(f.dest_uri, 'index.html')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/index.html')
                if use_directory_urls:
                    self.assertEqual(f.url, './')
                else:
                    self.assertEqual(f.url, 'index.html')
                self.assertEqual(f.name, 'index')
                self.assertTrue(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_md_readme_index_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('README.md', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'README.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/README.md')
                self.assertEqual(f.dest_uri, 'index.html')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/index.html')
                if use_directory_urls:
                    self.assertEqual(f.url, './')
                else:
                    self.assertEqual(f.url, 'index.html')
                self.assertEqual(f.name, 'index')
                self.assertTrue(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_md_index_file_nested(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/index.md', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/index.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/index.md')
                self.assertEqual(f.dest_uri, 'foo/index.html')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/index.html')
                if use_directory_urls:
                    self.assertEqual(f.url, 'foo/')
                else:
                    self.assertEqual(f.url, 'foo/index.html')
                self.assertEqual(f.name, 'index')
                self.assertTrue(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_static_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/bar.html', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/bar.html')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/bar.html')
                self.assertEqual(f.dest_uri, 'foo/bar.html')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar.html')
                self.assertEqual(f.url, 'foo/bar.html')
                self.assertEqual(f.name, 'bar')
                self.assertFalse(f.is_documentation_page())
                self.assertTrue(f.is_static_page())
                self.assertFalse(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_media_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/bar.jpg', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/bar.jpg')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/bar.jpg')
                self.assertEqual(f.dest_uri, 'foo/bar.jpg')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar.jpg')
                self.assertEqual(f.url, 'foo/bar.jpg')
                self.assertEqual(f.name, 'bar')
                self.assertFalse(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertTrue(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_javascript_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/bar.js', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/bar.js')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/bar.js')
                self.assertEqual(f.dest_uri, 'foo/bar.js')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar.js')
                self.assertEqual(f.url, 'foo/bar.js')
                self.assertEqual(f.name, 'bar')
                self.assertFalse(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertTrue(f.is_media_file())
                self.assertTrue(f.is_javascript())
                self.assertFalse(f.is_css())

    def test_css_file(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File('foo/bar.css', '/path/to/docs', '/path/to/site', use_directory_urls)
                self.assertEqual(f.src_uri, 'foo/bar.css')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo/bar.css')
                self.assertEqual(f.dest_uri, 'foo/bar.css')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo/bar.css')
                self.assertEqual(f.url, 'foo/bar.css')
                self.assertEqual(f.name, 'bar')
                self.assertFalse(f.is_documentation_page())
                self.assertFalse(f.is_static_page())
                self.assertTrue(f.is_media_file())
                self.assertFalse(f.is_javascript())
                self.assertTrue(f.is_css())

    def test_file_name_with_space(self):
        f = File('foo bar.md', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(f.src_uri, 'foo bar.md')
        self.assertPathsEqual(f.abs_src_path, '/path/to/docs/foo bar.md')
        self.assertEqual(f.dest_uri, 'foo bar.html')
        self.assertPathsEqual(f.abs_dest_path, '/path/to/site/foo bar.html')
        self.assertEqual(f.url, 'foo%20bar.html')
        self.assertEqual(f.name, 'foo bar')

    def test_file_name_with_custom_dest_uri(self):
        for use_directory_urls in True, False:
            with self.subTest(use_directory_urls=use_directory_urls):
                f = File(
                    'stuff/foo.md',
                    src_dir='/path/to/docs',
                    dest_dir='/path/to/site',
                    use_directory_urls=use_directory_urls,
                    dest_uri='stuff/1-foo/index.html',
                )
                self.assertEqual(f.src_uri, 'stuff/foo.md')
                self.assertPathsEqual(f.abs_src_path, '/path/to/docs/stuff/foo.md')
                self.assertEqual(f.dest_uri, 'stuff/1-foo/index.html')
                self.assertPathsEqual(f.abs_dest_path, '/path/to/site/stuff/1-foo/index.html')
                if use_directory_urls:
                    self.assertEqual(f.url, 'stuff/1-foo/')
                else:
                    self.assertEqual(f.url, 'stuff/1-foo/index.html')
                self.assertEqual(f.name, 'foo')

    def test_files(self):
        fs = [
            File('index.md', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.md', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.html', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.js', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.css', '/path/to/docs', '/path/to/site', use_directory_urls=True),
        ]
        files = Files(fs)
        self.assertEqual(list(files), fs)
        self.assertEqual(len(files), 6)
        self.assertEqual(files.documentation_pages(), [fs[0], fs[1]])
        self.assertEqual(files.static_pages(), [fs[2]])
        self.assertEqual(files.media_files(), [fs[3], fs[4], fs[5]])
        self.assertEqual(files.javascript_files(), [fs[4]])
        self.assertEqual(files.css_files(), [fs[5]])
        self.assertEqual(files.get_file_from_path('foo/bar.jpg'), fs[3])
        self.assertEqual(files.get_file_from_path('foo/bar.jpg'), fs[3])
        self.assertEqual(files.get_file_from_path('missing.jpg'), None)
        self.assertTrue(fs[2].src_uri in files.src_uris)
        extra_file = File('extra.md', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertFalse(extra_file.src_uri in files.src_uris)
        files.append(extra_file)
        self.assertEqual(len(files), 7)
        self.assertTrue(extra_file.src_uri in files.src_uris)
        self.assertEqual(files.documentation_pages(), [fs[0], fs[1], extra_file])
        files.remove(fs[1])
        self.assertEqual(files.documentation_pages(), [fs[0], extra_file])

    @tempdir(
        files=[
            'favicon.ico',
            'index.md',
        ]
    )
    @tempdir(
        files=[
            'base.html',
            'favicon.ico',
            'style.css',
            'foo.md',
            'README',
            '.ignore.txt',
            '.ignore/file.txt',
            'foo/.ignore.txt',
            'foo/.ignore/file.txt',
        ]
    )
    def test_add_files_from_theme(self, tdir, ddir):
        config = load_config(docs_dir=ddir, theme={'name': None, 'custom_dir': tdir})
        env = config.theme.get_env()
        files = get_files(config)
        self.assertEqual(
            [file.src_uri for file in files],
            ['index.md', 'favicon.ico'],
        )
        files.add_files_from_theme(env, config)
        self.assertEqual(
            [file.src_uri for file in files],
            ['index.md', 'favicon.ico', 'style.css'],
        )
        # Ensure theme file does not override docs_dir file
        self.assertEqual(
            files.get_file_from_path('favicon.ico').abs_src_path,
            os.path.normpath(os.path.join(ddir, 'favicon.ico')),
        )

    def test_get_relative_url_use_directory_urls(self):
        to_files = [
            'index.md',
            'foo/index.md',
            'foo/bar/index.md',
            'foo/bar/baz/index.md',
            'foo.md',
            'foo/bar.md',
            'foo/bar/baz.md',
        ]
        to_file_urls = [
            './',
            'foo/',
            'foo/bar/',
            'foo/bar/baz/',
            'foo/',
            'foo/bar/',
            'foo/bar/baz/',
        ]

        from_file = File('img.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertEqual(from_file.url, 'img.jpg')

        expected = [
            'img.jpg',  # img.jpg relative to .
            '../img.jpg',  # img.jpg relative to foo/
            '../../img.jpg',  # img.jpg relative to foo/bar/
            '../../../img.jpg',  # img.jpg relative to foo/bar/baz/
            '../img.jpg',  # img.jpg relative to foo
            '../../img.jpg',  # img.jpg relative to foo/bar
            '../../../img.jpg',  # img.jpg relative to foo/bar/baz
        ]
        for i, filename in enumerate(to_files):
            file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=True)
            self.assertEqual(file.url, to_file_urls[i])
            self.assertEqual(from_file.url_relative_to(file.url), expected[i])
            self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('foo/img.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertEqual(from_file.url, 'foo/img.jpg')

        expected = [
            'foo/img.jpg',  # foo/img.jpg relative to .
            'img.jpg',  # foo/img.jpg relative to foo/
            '../img.jpg',  # foo/img.jpg relative to foo/bar/
            '../../img.jpg',  # foo/img.jpg relative to foo/bar/baz/
            'img.jpg',  # foo/img.jpg relative to foo
            '../img.jpg',  # foo/img.jpg relative to foo/bar
            '../../img.jpg',  # foo/img.jpg relative to foo/bar/baz
        ]
        for i, filename in enumerate(to_files):
            file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=True)
            self.assertEqual(file.url, to_file_urls[i])
            self.assertEqual(from_file.url_relative_to(file.url), expected[i])
            self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('index.html', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertEqual(from_file.url, './')

        expected = [
            './',  # . relative to .
            '../',  # . relative to foo/
            '../../',  # . relative to foo/bar/
            '../../../',  # . relative to foo/bar/baz/
            '../',  # . relative to foo
            '../../',  # . relative to foo/bar
            '../../../',  # . relative to foo/bar/baz
        ]
        for i, filename in enumerate(to_files):
            file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=True)
            self.assertEqual(file.url, to_file_urls[i])
            self.assertEqual(from_file.url_relative_to(file.url), expected[i])
            self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('file.md', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertEqual(from_file.url, 'file/')

        expected = [
            'file/',  # file relative to .
            '../file/',  # file relative to foo/
            '../../file/',  # file relative to foo/bar/
            '../../../file/',  # file relative to foo/bar/baz/
            '../file/',  # file relative to foo
            '../../file/',  # file relative to foo/bar
            '../../../file/',  # file relative to foo/bar/baz
        ]
        for i, filename in enumerate(to_files):
            file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=True)
            self.assertEqual(file.url, to_file_urls[i])
            self.assertEqual(from_file.url_relative_to(file.url), expected[i])
            self.assertEqual(from_file.url_relative_to(file), expected[i])

    def test_get_relative_url(self):
        to_files = [
            'index.md',
            'foo/index.md',
            'foo/bar/index.md',
            'foo/bar/baz/index.md',
            'foo.md',
            'foo/bar.md',
            'foo/bar/baz.md',
        ]
        to_file_urls = [
            'index.html',
            'foo/index.html',
            'foo/bar/index.html',
            'foo/bar/baz/index.html',
            'foo.html',
            'foo/bar.html',
            'foo/bar/baz.html',
        ]

        from_file = File('img.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(from_file.url, 'img.jpg')

        expected = [
            'img.jpg',  # img.jpg relative to .
            '../img.jpg',  # img.jpg relative to foo/
            '../../img.jpg',  # img.jpg relative to foo/bar/
            '../../../img.jpg',  # img.jpg relative to foo/bar/baz/
            'img.jpg',  # img.jpg relative to foo.html
            '../img.jpg',  # img.jpg relative to foo/bar.html
            '../../img.jpg',  # img.jpg relative to foo/bar/baz.html
        ]
        for i, filename in enumerate(to_files):
            with self.subTest(from_file=from_file.src_path, to_file=filename):
                file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=False)
                self.assertEqual(file.url, to_file_urls[i])
                self.assertEqual(from_file.url_relative_to(file.url), expected[i])
                self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('foo/img.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(from_file.url, 'foo/img.jpg')

        expected = [
            'foo/img.jpg',  # foo/img.jpg relative to .
            'img.jpg',  # foo/img.jpg relative to foo/
            '../img.jpg',  # foo/img.jpg relative to foo/bar/
            '../../img.jpg',  # foo/img.jpg relative to foo/bar/baz/
            'foo/img.jpg',  # foo/img.jpg relative to foo.html
            'img.jpg',  # foo/img.jpg relative to foo/bar.html
            '../img.jpg',  # foo/img.jpg relative to foo/bar/baz.html
        ]
        for i, filename in enumerate(to_files):
            with self.subTest(from_file=from_file.src_path, to_file=filename):
                file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=False)
                self.assertEqual(file.url, to_file_urls[i])
                self.assertEqual(from_file.url_relative_to(file.url), expected[i])
                self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('index.html', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(from_file.url, 'index.html')

        expected = [
            'index.html',  # index.html relative to .
            '../index.html',  # index.html relative to foo/
            '../../index.html',  # index.html relative to foo/bar/
            '../../../index.html',  # index.html relative to foo/bar/baz/
            'index.html',  # index.html relative to foo.html
            '../index.html',  # index.html relative to foo/bar.html
            '../../index.html',  # index.html relative to foo/bar/baz.html
        ]
        for i, filename in enumerate(to_files):
            with self.subTest(from_file=from_file.src_path, to_file=filename):
                file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=False)
                self.assertEqual(file.url, to_file_urls[i])
                self.assertEqual(from_file.url_relative_to(file.url), expected[i])
                self.assertEqual(from_file.url_relative_to(file), expected[i])

        from_file = File('file.html', '/path/to/docs', '/path/to/site', use_directory_urls=False)
        self.assertEqual(from_file.url, 'file.html')

        expected = [
            'file.html',  # file.html relative to .
            '../file.html',  # file.html relative to foo/
            '../../file.html',  # file.html relative to foo/bar/
            '../../../file.html',  # file.html relative to foo/bar/baz/
            'file.html',  # file.html relative to foo.html
            '../file.html',  # file.html relative to foo/bar.html
            '../../file.html',  # file.html relative to foo/bar/baz.html
        ]
        for i, filename in enumerate(to_files):
            with self.subTest(from_file=from_file.src_path, to_file=filename):
                file = File(filename, '/path/to/docs', '/path/to/site', use_directory_urls=False)
                self.assertEqual(file.url, to_file_urls[i])
                self.assertEqual(from_file.url_relative_to(file.url), expected[i])
                self.assertEqual(from_file.url_relative_to(file), expected[i])

    @tempdir(
        files=[
            'index.md',
            'readme.md',
            'bar.css',
            'bar.html',
            'bar.jpg',
            'bar.js',
            'bar.md',
            '.dotfile',
            'templates/foo.html',
        ]
    )
    def test_get_files(self, tdir):
        config = load_config(docs_dir=tdir, extra_css=['bar.css'], extra_javascript=['bar.js'])
        files = get_files(config)
        self.assertIsInstance(files, Files)
        self.assertEqual(
            [f.src_uri for f in files if f.inclusion.is_included()],
            ['index.md', 'bar.css', 'bar.html', 'bar.jpg', 'bar.js', 'bar.md', 'readme.md'],
        )
        self.assertEqual(
            [f.src_uri for f in files if f.inclusion.is_excluded()],
            ['.dotfile', 'templates/foo.html'],
        )

    @tempdir(
        files=[
            'README.md',
            'foo.md',
        ]
    )
    def test_get_files_include_readme_without_index(self, tdir):
        config = load_config(docs_dir=tdir)
        files = get_files(config)
        self.assertIsInstance(files, Files)
        self.assertEqual([f.src_uri for f in files], ['README.md', 'foo.md'])

    @tempdir(
        files=[
            'index.md',
            'README.md',
            'foo.md',
        ]
    )
    def test_get_files_exclude_readme_with_index(self, tdir):
        config = load_config(docs_dir=tdir)
        with self.assertLogs('mkdocs') as cm:
            files = get_files(config)
        self.assertRegex(
            '\n'.join(cm.output),
            r"^WARNING:mkdocs.structure.files:"
            r"Excluding 'README.md' from the site because it conflicts with 'index.md'.$",
        )
        self.assertIsInstance(files, Files)
        self.assertEqual([f.src_uri for f in files], ['index.md', 'foo.md'])

    @tempdir()
    @tempdir(files={'test.txt': 'source content'})
    def test_copy_file(self, src_dir, dest_dir):
        file = File('test.txt', src_dir, dest_dir, use_directory_urls=False)
        dest_path = os.path.join(dest_dir, 'test.txt')
        self.assertPathNotExists(dest_path)
        file.copy_file()
        self.assertPathIsFile(dest_path)

    @tempdir(files={'test.txt': 'source content'})
    def test_copy_file_same_file(self, dest_dir):
        file = File('test.txt', dest_dir, dest_dir, use_directory_urls=False)
        dest_path = os.path.join(dest_dir, 'test.txt')
        file.copy_file()
        self.assertPathIsFile(dest_path)
        with open(dest_path, encoding='utf-8') as f:
            self.assertEqual(f.read(), 'source content')

    @tempdir(files={'test.txt': 'destination content'})
    @tempdir(files={'test.txt': 'source content'})
    def test_copy_file_clean_modified(self, src_dir, dest_dir):
        file = File('test.txt', src_dir, dest_dir, use_directory_urls=False)
        file.is_modified = mock.Mock(return_value=True)
        dest_path = os.path.join(dest_dir, 'test.txt')
        file.copy_file(dirty=False)
        self.assertPathIsFile(dest_path)
        with open(dest_path, encoding='utf-8') as f:
            self.assertEqual(f.read(), 'source content')

    @tempdir(files={'test.txt': 'destination content'})
    @tempdir(files={'test.txt': 'source content'})
    def test_copy_file_dirty_modified(self, src_dir, dest_dir):
        file = File('test.txt', src_dir, dest_dir, use_directory_urls=False)
        file.is_modified = mock.Mock(return_value=True)
        dest_path = os.path.join(dest_dir, 'test.txt')
        file.copy_file(dirty=True)
        self.assertPathIsFile(dest_path)
        with open(dest_path, encoding='utf-8') as f:
            self.assertEqual(f.read(), 'source content')

    @tempdir(files={'test.txt': 'destination content'})
    @tempdir(files={'test.txt': 'source content'})
    def test_copy_file_dirty_not_modified(self, src_dir, dest_dir):
        file = File('test.txt', src_dir, dest_dir, use_directory_urls=False)
        file.is_modified = mock.Mock(return_value=False)
        dest_path = os.path.join(dest_dir, 'test.txt')
        file.copy_file(dirty=True)
        self.assertPathIsFile(dest_path)
        with open(dest_path, encoding='utf-8') as f:
            self.assertEqual(f.read(), 'destination content')

    def test_files_append_remove_src_paths(self):
        fs = [
            File('index.md', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.md', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.html', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.jpg', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.js', '/path/to/docs', '/path/to/site', use_directory_urls=True),
            File('foo/bar.css', '/path/to/docs', '/path/to/site', use_directory_urls=True),
        ]
        files = Files(fs)
        self.assertEqual(len(files), 6)
        self.assertEqual(len(files.src_uris), 6)
        extra_file = File('extra.md', '/path/to/docs', '/path/to/site', use_directory_urls=True)
        self.assertFalse(extra_file.src_uri in files.src_uris)
        files.append(extra_file)
        self.assertEqual(len(files), 7)
        self.assertEqual(len(files.src_uris), 7)
        self.assertTrue(extra_file.src_uri in files.src_uris)
        files.remove(extra_file)
        self.assertEqual(len(files), 6)
        self.assertEqual(len(files.src_uris), 6)
        self.assertFalse(extra_file.src_uri in files.src_uris)



from __future__ import annotations

from click import ClickException, echo


class MkDocsException(ClickException):
    """
    The base class which all MkDocs exceptions inherit from. This should
    not be raised directly. One of the subclasses should be raised instead.
    """


class Abort(MkDocsException, SystemExit):
    """Abort the build."""

    code = 1

    def show(self, *args, **kwargs) -> None:
        echo('\n' + self.format_message())


class ConfigurationError(MkDocsException):
    """
    This error is raised by configuration validation when a validation error
    is encountered. This error should be raised by any configuration options
    defined in a plugin's [config_scheme][].
    """


class BuildError(MkDocsException):
    """
    This error may be raised by MkDocs during the build process. Plugins should
    not raise this error.
    """


class PluginError(BuildError):
    """
    A subclass of [`mkdocs.exceptions.BuildError`][] which can be raised by plugin
    events.
    """



#!/usr/bin/env python
from __future__ import annotations

import contextlib
import io
import os.path
import re
import textwrap
import unittest
from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import markdown.preprocessors

from mkdocs.commands import build
from mkdocs.config import base
from mkdocs.exceptions import PluginError
from mkdocs.structure.files import File, Files
from mkdocs.structure.nav import get_navigation
from mkdocs.structure.pages import Page
from mkdocs.tests.base import PathAssertionMixin, load_config, tempdir
from mkdocs.utils import meta

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig


def build_page(title, path, config, md_src=''):
    """Helper which returns a Page object."""
    files = Files([File(path, config.docs_dir, config.site_dir, config.use_directory_urls)])
    page = Page(title, list(files)[0], config)
    # Fake page.read_source()
    page.markdown, page.meta = meta.get_data(md_src)
    return page, files


class BuildTests(PathAssertionMixin, unittest.TestCase):
    def _get_env_with_null_translations(self, config):
        env = config.theme.get_env()
        env.add_extension('jinja2.ext.i18n')
        env.install_null_translations()
        return env

    # Test build.get_context

    def test_context_base_url_homepage(self):
        nav_cfg = [
            {'Home': 'index.md'},
        ]
        cfg = load_config(nav=nav_cfg, use_directory_urls=False)
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[0])
        self.assertEqual(context['base_url'], '.')

    def test_context_base_url_homepage_use_directory_urls(self):
        nav_cfg = [
            {'Home': 'index.md'},
        ]
        cfg = load_config(nav=nav_cfg)
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[0])
        self.assertEqual(context['base_url'], '.')

    def test_context_base_url_nested_page(self):
        nav_cfg = [
            {'Home': 'index.md'},
            {'Nested': 'foo/bar.md'},
        ]
        cfg = load_config(nav=nav_cfg, use_directory_urls=False)
        fs = [
            File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
            File('foo/bar.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
        ]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[1])
        self.assertEqual(context['base_url'], '..')

    def test_context_base_url_nested_page_use_directory_urls(self):
        nav_cfg = [
            {'Home': 'index.md'},
            {'Nested': 'foo/bar.md'},
        ]
        cfg = load_config(nav=nav_cfg)
        fs = [
            File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
            File('foo/bar.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
        ]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[1])
        self.assertEqual(context['base_url'], '../..')

    def test_context_base_url_relative_no_page(self):
        cfg = load_config(use_directory_urls=False)
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='..')
        self.assertEqual(context['base_url'], '..')

    def test_context_base_url_relative_no_page_use_directory_urls(self):
        cfg = load_config()
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='..')
        self.assertEqual(context['base_url'], '..')

    def test_context_base_url_absolute_no_page(self):
        cfg = load_config(use_directory_urls=False)
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='/')
        self.assertEqual(context['base_url'], '/')

    def test_context_base_url__absolute_no_page_use_directory_urls(self):
        cfg = load_config()
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='/')
        self.assertEqual(context['base_url'], '/')

    def test_context_base_url_absolute_nested_no_page(self):
        cfg = load_config(use_directory_urls=False)
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='/foo/')
        self.assertEqual(context['base_url'], '/foo/')

    def test_context_base_url__absolute_nested_no_page_use_directory_urls(self):
        cfg = load_config()
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='/foo/')
        self.assertEqual(context['base_url'], '/foo/')

    def test_context_extra_css_js_from_homepage(self):
        nav_cfg = [
            {'Home': 'index.md'},
        ]
        cfg = load_config(
            nav=nav_cfg,
            extra_css=['style.css'],
            extra_javascript=['script.js'],
            use_directory_urls=False,
        )
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[0])
        self.assertEqual(context['extra_css'], ['style.css'])
        self.assertEqual(context['extra_javascript'], ['script.js'])

    def test_context_extra_css_js_from_nested_page(self):
        nav_cfg = [
            {'Home': 'index.md'},
            {'Nested': 'foo/bar.md'},
        ]
        cfg = load_config(
            nav=nav_cfg,
            extra_css=['style.css'],
            extra_javascript=['script.js'],
            use_directory_urls=False,
        )
        fs = [
            File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
            File('foo/bar.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
        ]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[1])
        self.assertEqual(context['extra_css'], ['../style.css'])
        self.assertEqual(context['extra_javascript'], ['../script.js'])

    def test_context_extra_css_js_from_nested_page_use_directory_urls(self):
        nav_cfg = [
            {'Home': 'index.md'},
            {'Nested': 'foo/bar.md'},
        ]
        cfg = load_config(
            nav=nav_cfg,
            extra_css=['style.css'],
            extra_javascript=['script.js'],
        )
        fs = [
            File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
            File('foo/bar.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls),
        ]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        context = build.get_context(nav, files, cfg, nav.pages[1])
        self.assertEqual(context['extra_css'], ['../../style.css'])
        self.assertEqual(context['extra_javascript'], ['../../script.js'])

    # TODO: This shouldn't pass on Linux
    # @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_context_extra_css_path_warning(self):
        nav_cfg = [
            {'Home': 'index.md'},
        ]
        cfg = load_config(
            nav=nav_cfg,
            extra_css=['assets\\style.css'],
            use_directory_urls=False,
        )
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        with self.assertLogs('mkdocs') as cm:
            context = build.get_context(nav, files, cfg, nav.pages[0])
        self.assertEqual(context['extra_css'], ['assets/style.css'])
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.utils:Path 'assets\\style.css' uses OS-specific separator '\\'. "
            "That will be unsupported in a future release. Please change it to '/'.",
        )

    def test_context_extra_css_js_no_page(self):
        cfg = load_config(extra_css=['style.css'], extra_javascript=['script.js'])
        context = build.get_context(mock.Mock(), mock.Mock(), cfg, base_url='..')
        self.assertEqual(context['extra_css'], ['../style.css'])
        self.assertEqual(context['extra_javascript'], ['../script.js'])

    def test_extra_context(self):
        cfg = load_config(extra={'a': 1})
        context = build.get_context(mock.Mock(), mock.Mock(), cfg)
        self.assertEqual(context['config']['extra']['a'], 1)

    # Test build._build_theme_template

    @mock.patch('mkdocs.utils.write_file')
    @mock.patch('mkdocs.commands.build._build_template', return_value='some content')
    def test_build_theme_template(self, mock_build_template, mock_write_file):
        cfg = load_config()
        env = cfg.theme.get_env()
        build._build_theme_template('main.html', env, Files([]), cfg, mock.Mock())
        mock_write_file.assert_called_once()
        mock_build_template.assert_called_once()

    @mock.patch('mkdocs.utils.write_file')
    @mock.patch('mkdocs.commands.build._build_template', return_value='some content')
    @mock.patch('gzip.GzipFile')
    @tempdir()
    def test_build_sitemap_template(
        self, site_dir, mock_gzip_gzipfile, mock_build_template, mock_write_file
    ):
        cfg = load_config(site_dir=site_dir)
        env = cfg.theme.get_env()
        build._build_theme_template('sitemap.xml', env, Files([]), cfg, mock.Mock())
        mock_write_file.assert_called_once()
        mock_build_template.assert_called_once()
        mock_gzip_gzipfile.assert_called_once()

    @mock.patch('mkdocs.utils.write_file')
    @mock.patch('mkdocs.commands.build._build_template', return_value='')
    def test_skip_missing_theme_template(self, mock_build_template, mock_write_file):
        cfg = load_config()
        env = cfg.theme.get_env()
        with self.assertLogs('mkdocs') as cm:
            build._build_theme_template('missing.html', env, Files([]), cfg, mock.Mock())
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.commands.build:Template skipped: 'missing.html' not found in theme directories.",
        )
        mock_write_file.assert_not_called()
        mock_build_template.assert_not_called()

    @mock.patch('mkdocs.utils.write_file')
    @mock.patch('mkdocs.commands.build._build_template', return_value='')
    def test_skip_theme_template_empty_output(self, mock_build_template, mock_write_file):
        cfg = load_config()
        env = cfg.theme.get_env()
        with self.assertLogs('mkdocs') as cm:
            build._build_theme_template('main.html', env, Files([]), cfg, mock.Mock())
        self.assertEqual(
            '\n'.join(cm.output),
            "INFO:mkdocs.commands.build:Template skipped: 'main.html' generated empty output.",
        )
        mock_write_file.assert_not_called()
        mock_build_template.assert_called_once()

    # Test build._build_extra_template

    @tempdir()
    @mock.patch('mkdocs.commands.build.open', mock.mock_open(read_data='template content'))
    def test_build_extra_template(self, site_dir):
        cfg = load_config(site_dir=site_dir)
        fs = [File('foo.html', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        build._build_extra_template('foo.html', files, cfg, mock.Mock())

    @mock.patch('mkdocs.commands.build.open', mock.mock_open(read_data='template content'))
    def test_skip_missing_extra_template(self):
        cfg = load_config()
        fs = [File('foo.html', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        with self.assertLogs('mkdocs') as cm:
            build._build_extra_template('missing.html', files, cfg, mock.Mock())
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.commands.build:Template skipped: 'missing.html' not found in docs_dir.",
        )

    @mock.patch('mkdocs.commands.build.open', mock.Mock(side_effect=OSError('Error message.')))
    def test_skip_ioerror_extra_template(self):
        cfg = load_config()
        fs = [File('foo.html', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        with self.assertLogs('mkdocs') as cm:
            build._build_extra_template('foo.html', files, cfg, mock.Mock())
        self.assertEqual(
            '\n'.join(cm.output),
            "WARNING:mkdocs.commands.build:Error reading template 'foo.html': Error message.",
        )

    @mock.patch('mkdocs.commands.build.open', mock.mock_open(read_data=''))
    def test_skip_extra_template_empty_output(self):
        cfg = load_config()
        fs = [File('foo.html', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        with self.assertLogs('mkdocs') as cm:
            build._build_extra_template('foo.html', files, cfg, mock.Mock())
        self.assertEqual(
            '\n'.join(cm.output),
            "INFO:mkdocs.commands.build:Template skipped: 'foo.html' generated empty output.",
        )

    # Test build._populate_page

    @tempdir(files={'index.md': 'page content'})
    def test_populate_page(self, docs_dir):
        cfg = load_config(docs_dir=docs_dir)
        file = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        page = Page('Foo', file, cfg)
        build._populate_page(page, cfg, Files([file]))
        self.assertEqual(page.content, '<p>page content</p>')

    @tempdir(files={'testing.html': '<p>page content</p>'})
    def test_populate_page_dirty_modified(self, site_dir):
        cfg = load_config(site_dir=site_dir)
        file = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        page = Page('Foo', file, cfg)
        build._populate_page(page, cfg, Files([file]), dirty=True)
        self.assertTrue(page.markdown.startswith('# Welcome to MkDocs'))
        self.assertTrue(
            page.content.startswith('<h1 id="welcome-to-mkdocs">Welcome to MkDocs</h1>')
        )

    @tempdir(files={'index.md': 'page content'})
    @tempdir(files={'index.html': '<p>page content</p>'})
    def test_populate_page_dirty_not_modified(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir)
        file = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        page = Page('Foo', file, cfg)
        build._populate_page(page, cfg, Files([file]), dirty=True)
        # Content is empty as file read was skipped
        self.assertEqual(page.markdown, None)
        self.assertEqual(page.content, None)

    @tempdir(files={'index.md': 'new page content'})
    @mock.patch('mkdocs.structure.pages.open', side_effect=OSError('Error message.'))
    def test_populate_page_read_error(self, docs_dir, mock_open):
        cfg = load_config(docs_dir=docs_dir)
        file = File('missing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        page = Page('Foo', file, cfg)
        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(OSError):
                build._populate_page(page, cfg, Files([file]))
        self.assertEqual(
            cm.output,
            [
                'ERROR:mkdocs.structure.pages:File not found: missing.md',
                "ERROR:mkdocs.commands.build:Error reading page 'missing.md': Error message.",
            ],
        )
        mock_open.assert_called_once()

    @tempdir(files={'index.md': 'page content'})
    def test_populate_page_read_plugin_error(self, docs_dir):
        def on_page_markdown(*args, **kwargs):
            raise PluginError('Error message.')

        cfg = load_config(docs_dir=docs_dir)
        cfg.plugins.events['page_markdown'].append(on_page_markdown)

        file = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        page = Page('Foo', file, cfg)
        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(PluginError):
                build._populate_page(page, cfg, Files([file]))
        self.assertEqual(
            '\n'.join(cm.output),
            "ERROR:mkdocs.commands.build:Error reading page 'index.md':",
        )

    # Test build._build_page

    @tempdir()
    def test_build_page(self, site_dir):
        cfg = load_config(site_dir=site_dir, nav=['index.md'])
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.markdown = 'page content'
        page.content = '<p>page content</p>'
        build._build_page(page, cfg, files, nav, self._get_env_with_null_translations(cfg))
        self.assertPathIsFile(site_dir, 'index.html')

    @tempdir()
    @mock.patch('jinja2.environment.Template.render', return_value='')
    def test_build_page_empty(self, site_dir, render_mock):
        cfg = load_config(site_dir=site_dir, nav=['index.md'])
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        with self.assertLogs('mkdocs') as cm:
            build._build_page(
                files.documentation_pages()[0].page, cfg, files, nav, cfg.theme.get_env()
            )
        self.assertEqual(
            '\n'.join(cm.output),
            "INFO:mkdocs.commands.build:Page skipped: 'index.md'. Generated empty output.",
        )
        self.assertPathNotExists(site_dir, 'index.html')
        render_mock.assert_called_once()

    @tempdir(files={'index.md': 'page content'})
    @tempdir(files={'index.html': '<p>page content</p>'})
    @mock.patch('mkdocs.utils.write_file')
    def test_build_page_dirty_modified(self, site_dir, docs_dir, mock_write_file):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir, nav=['index.md'])
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.markdown = 'new page content'
        page.content = '<p>new page content</p>'
        build._build_page(
            page, cfg, files, nav, self._get_env_with_null_translations(cfg), dirty=True
        )
        mock_write_file.assert_not_called()

    @tempdir(files={'testing.html': '<p>page content</p>'})
    @mock.patch('mkdocs.utils.write_file')
    def test_build_page_dirty_not_modified(self, site_dir, mock_write_file):
        cfg = load_config(site_dir=site_dir, nav=['testing.md'])
        fs = [File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.markdown = 'page content'
        page.content = '<p>page content</p>'
        build._build_page(
            page, cfg, files, nav, self._get_env_with_null_translations(cfg), dirty=True
        )
        mock_write_file.assert_called_once()

    @tempdir()
    def test_build_page_custom_template(self, site_dir):
        cfg = load_config(site_dir=site_dir, nav=['index.md'])
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.meta = {'template': '404.html'}
        page.markdown = 'page content'
        page.content = '<p>page content</p>'
        build._build_page(page, cfg, files, nav, self._get_env_with_null_translations(cfg))
        self.assertPathIsFile(site_dir, 'index.html')

    @tempdir()
    @mock.patch('mkdocs.utils.write_file', side_effect=OSError('Error message.'))
    def test_build_page_error(self, site_dir, mock_write_file):
        cfg = load_config(site_dir=site_dir, nav=['index.md'])
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.markdown = 'page content'
        page.content = '<p>page content</p>'
        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(OSError):
                build._build_page(page, cfg, files, nav, self._get_env_with_null_translations(cfg))
        self.assertEqual(
            '\n'.join(cm.output),
            "ERROR:mkdocs.commands.build:Error building page 'index.md': Error message.",
        )
        mock_write_file.assert_called_once()

    @tempdir()
    def test_build_page_plugin_error(self, site_dir):
        def on_page_context(*args, **kwargs):
            raise PluginError('Error message.')

        cfg = load_config(site_dir=site_dir, nav=['index.md'])
        cfg.plugins.events['page_context'].append(on_page_context)
        fs = [File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)]
        files = Files(fs)
        nav = get_navigation(files, cfg)
        page = files.documentation_pages()[0].page
        # Fake populate page
        page.title = 'Title'
        page.markdown = 'page content'
        page.content = '<p>page content</p>'
        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(PluginError):
                build._build_page(page, cfg, files, nav, cfg.theme.get_env())
        self.assertEqual(
            '\n'.join(cm.output),
            "ERROR:mkdocs.commands.build:Error building page 'index.md':",
        )

    # Test build.build

    @tempdir(
        files={
            'index.md': 'page content',
            'empty.md': '',
            'img.jpg': '',
            'static.html': 'content',
            '.hidden': 'content',
            '.git/hidden': 'content',
        }
    )
    @tempdir()
    def test_copying_media(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir)
        build.build(cfg)

        # Verify that only non-empty md file (converted to html), static HTML file and image are copied.
        self.assertPathIsFile(site_dir, 'index.html')
        self.assertPathIsFile(site_dir, 'img.jpg')
        self.assertPathIsFile(site_dir, 'static.html')
        self.assertPathNotExists(site_dir, 'empty.md')
        self.assertPathNotExists(site_dir, '.hidden')
        self.assertPathNotExists(site_dir, '.git/hidden')

    @tempdir(files={'index.md': 'page content'})
    @tempdir()
    def test_copy_theme_files(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir)
        build.build(cfg)

        # Verify only theme media are copied, not templates, Python or localization files.
        self.assertPathIsFile(site_dir, 'index.html')
        self.assertPathIsFile(site_dir, '404.html')
        self.assertPathIsDir(site_dir, 'js')
        self.assertPathIsDir(site_dir, 'css')
        self.assertPathIsDir(site_dir, 'img')
        self.assertPathIsDir(site_dir, 'fonts')
        self.assertPathNotExists(site_dir, '__init__.py')
        self.assertPathNotExists(site_dir, '__init__.pyc')
        self.assertPathNotExists(site_dir, 'base.html')
        self.assertPathNotExists(site_dir, 'content.html')
        self.assertPathNotExists(site_dir, 'main.html')
        self.assertPathNotExists(site_dir, 'locales')

    @contextlib.contextmanager
    def _assert_build_logs(self, expected):
        with self.assertLogs('mkdocs') as cm:
            yield
        msgs = [f'{r.levelname}:{r.message}' for r in cm.records]
        if msgs and msgs[0].startswith('INFO:Cleaning site directory'):
            del msgs[0]
        if msgs and msgs[0].startswith('INFO:Building documentation to directory'):
            del msgs[0]
        if msgs and msgs[-1].startswith('INFO:Documentation built'):
            del msgs[-1]
        self.assertEqual('\n'.join(msgs), textwrap.dedent(expected).strip('\n'))

    @tempdir(
        files={
            'test/foo.md': 'page1 content, [bar](bar.md)',
            'test/bar.md': 'page2 content, [baz](baz.md), [nonexistent](nonexistent.md)',
            'test/baz.md': 'page3 content, [foo](foo.md)',
            '.zoo.md': 'page4 content',
        }
    )
    @tempdir()
    def test_draft_pages_with_invalid_links(self, site_dir, docs_dir):
        cfg = load_config(
            docs_dir=docs_dir,
            site_dir=site_dir,
            use_directory_urls=False,
            draft_docs='ba*.md',
        )

        with self.subTest(serve_url=None):
            expected_logs = '''
                INFO:Doc file 'test/foo.md' contains a link to 'test/bar.md' which is excluded from the built site.
            '''
            with self._assert_build_logs(expected_logs):
                build.build(cfg)
            self.assertPathIsFile(site_dir, 'test', 'foo.html')
            self.assertPathNotExists(site_dir, 'test', 'baz.html')
            self.assertPathNotExists(site_dir, '.zoo.html')

        serve_url = 'http://localhost:123/documentation/'
        with self.subTest(serve_url=serve_url):
            expected_logs = '''
                INFO:Doc file 'test/bar.md' contains a link 'nonexistent.md', but the target 'test/nonexistent.md' is not found among documentation files.
                INFO:Doc file 'test/foo.md' contains a link to 'test/bar.md' which is excluded from the built site.
                INFO:The following pages are being built only for the preview but will be excluded from `mkdocs build` per `draft_docs` config:
                  - http://localhost:123/documentation/test/bar.html
                  - http://localhost:123/documentation/test/baz.html
            '''
            with self._assert_build_logs(expected_logs):
                build.build(cfg, serve_url=serve_url)

            foo_path = Path(site_dir, 'test', 'foo.html')
            self.assertTrue(foo_path.is_file())
            self.assertNotIn('DRAFT', foo_path.read_text())

            baz_path = Path(site_dir, 'test', 'baz.html')
            self.assertPathIsFile(baz_path)
            self.assertIn('DRAFT', baz_path.read_text())

            self.assertPathNotExists(site_dir, '.zoo.html')

    @tempdir(
        files={
            'foo/README.md': 'page1 content',
            'foo/index.md': 'page2 content',
        }
    )
    @tempdir()
    def test_conflicting_readme_and_index(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir, use_directory_urls=False)

        for serve_url in None, 'http://localhost:123/':
            with self.subTest(serve_url=serve_url):
                expected_logs = '''
                    WARNING:Excluding 'foo/README.md' from the site because it conflicts with 'foo/index.md'.
                '''
                with self._assert_build_logs(expected_logs):
                    build.build(cfg, serve_url=serve_url)

                index_path = Path(site_dir, 'foo', 'index.html')
                self.assertPathIsFile(index_path)
                self.assertRegex(index_path.read_text(), r'page2 content')

    @tempdir(
        files={
            'foo/README.md': 'page1 content',
            'foo/index.md': 'page2 content',
        }
    )
    @tempdir()
    def test_exclude_readme_and_index(self, site_dir, docs_dir):
        cfg = load_config(
            docs_dir=docs_dir, site_dir=site_dir, use_directory_urls=False, exclude_docs='index.md'
        )

        for serve_url in None, 'http://localhost:123/':
            with self.subTest(serve_url=serve_url):
                with self._assert_build_logs(''):
                    build.build(cfg, serve_url=serve_url)

                index_path = Path(site_dir, 'foo', 'index.html')
                self.assertPathIsFile(index_path)
                self.assertRegex(index_path.read_text(), r'page1 content')

    @tempdir(
        files={
            'test/foo.md': '## page1 heading\n\n[bar](bar.md#page1-heading)',
            'test/bar.md': '## page2 heading\n\n[aaa](#a)',
        }
    )
    @tempdir()
    def test_anchor_warning(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir, validation={'anchors': 'warn'})

        expected_logs = '''
            WARNING:Doc file 'test/bar.md' contains a link '#a', but there is no such anchor on this page.
            WARNING:Doc file 'test/foo.md' contains a link 'bar.md#page1-heading', but the doc 'test/bar.md' does not contain an anchor '#page1-heading'.
        '''
        with self._assert_build_logs(expected_logs):
            build.build(cfg)

    @tempdir(
        files={
            'test/foo.md': '[bar](bar.md#heading1)',
            'test/bar.md': '## heading1',
        }
    )
    @tempdir()
    def test_anchor_no_warning(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir, validation={'anchors': 'warn'})
        build.build(cfg)

    @tempdir(
        files={
            'test/foo.md': '[bar](bar.md#heading2)',
            'test/bar.md': '## heading1\n\n<a id="heading2">hi</a>',
        }
    )
    @tempdir()
    def test_anchor_no_warning_with_html(self, site_dir, docs_dir):
        cfg = load_config(
            docs_dir=docs_dir,
            site_dir=site_dir,
            validation={'anchors': 'warn'},
            markdown_extensions=['md_in_html'],
        )
        build.build(cfg)

    @tempdir(
        files={
            'test/foo.md': '[bar1](bar.md?test#heading1), [bar2](bar.md?test#headingno), [bar3](bar.md?test), [self4](?test), [self5](?test#hi)',
            'test/bar.md': '## heading1',
        }
    )
    @tempdir()
    def test_anchor_warning_and_query(self, site_dir, docs_dir):
        cfg = load_config(docs_dir=docs_dir, site_dir=site_dir, validation={'anchors': 'info'})

        expected_logs = '''
            INFO:Doc file 'test/foo.md' contains a link 'bar.md?test#headingno', but the doc 'test/bar.md' does not contain an anchor '#headingno'.
            INFO:Doc file 'test/foo.md' contains a link '?test#hi', but there is no such anchor on this page.
        '''
        with self._assert_build_logs(expected_logs):
            build.build(cfg)

    @tempdir(
        files={
            'test/foo.md': 'test[^1]\n\n[^1]: footnote1\n[^2]: footnote2',
        }
    )
    @tempdir()
    def test_anchor_warning_for_footnote(self, site_dir, docs_dir):
        cfg = load_config(
            docs_dir=docs_dir,
            site_dir=site_dir,
            validation={'anchors': 'info'},
            markdown_extensions=['footnotes'],
        )

        expected_logs = '''
            INFO:Doc file 'test/foo.md' contains a link '#fnref:2', but there is no such anchor on this page. This seems to be a footnote that is never referenced.
        '''
        with self._assert_build_logs(expected_logs):
            build.build(cfg)

    @tempdir(
        files={
            'foo.md': 'page1 content',
            'bar.md': 'page2 content',
        }
    )
    @tempdir()
    @tempdir()
    def test_plugins_adding_files_and_interacting(self, tmp_dir, site_dir, docs_dir):
        def on_files_1(files: Files, config: MkDocsConfig) -> Files:
            # Plugin 1 generates a file.
            Path(tmp_dir, 'SUMMARY.md').write_text('foo.md\nbar.md\n')
            files.append(File('SUMMARY.md', tmp_dir, config.site_dir, config.use_directory_urls))
            return files

        def on_files_2(files: Files, config: MkDocsConfig) -> None:
            # Plugin 2 reads that file and uses it to configure the nav.
            f = files.get_file_from_path('SUMMARY.md')
            assert f is not None
            config.nav = Path(f.abs_src_path).read_text().splitlines()

        for serve_url in None, 'http://localhost:123/':
            for exclude in 'full', 'drafts', 'nav', None:
                with self.subTest(serve_url=serve_url, exclude=exclude):
                    cfg = load_config(
                        docs_dir=docs_dir,
                        site_dir=site_dir,
                        use_directory_urls=False,
                        exclude_docs='SUMMARY.md' if exclude == 'full' else '',
                        draft_docs='SUMMARY.md' if exclude == 'drafts' else '',
                        not_in_nav='SUMMARY.md' if exclude == 'nav' else '',
                    )
                    cfg.plugins.events['files'] += [on_files_1, on_files_2]

                    expected_logs = ''
                    if exclude is None:
                        expected_logs = '''
                            INFO:The following pages exist in the docs directory, but are not included in the "nav" configuration:
                              - SUMMARY.md
                        '''
                    if exclude == 'drafts' and serve_url:
                        expected_logs = '''
                            INFO:The following pages are being built only for the preview but will be excluded from `mkdocs build` per `draft_docs` config:
                              - http://localhost:123/SUMMARY.html
                        '''
                    with self._assert_build_logs(expected_logs):
                        build.build(cfg, serve_url=serve_url)

                    foo_path = Path(site_dir, 'foo.html')
                    self.assertPathIsFile(foo_path)
                    self.assertRegex(
                        foo_path.read_text(),
                        r'href="foo.html"[\s\S]+href="bar.html"',  # Nav order is respected
                    )

                    summary_path = Path(site_dir, 'SUMMARY.html')
                    if exclude == 'full':
                        self.assertPathNotExists(summary_path)
                    elif exclude == 'drafts' and not serve_url:
                        self.assertPathNotExists(summary_path)
                    else:
                        self.assertPathExists(summary_path)

    @tempdir(
        files={
            'README.md': 'CONFIG_README\n',
            'docs/foo.md': 'ROOT_FOO\n',
            'docs/test/bar.md': 'TEST_BAR\n',
            'docs/main/foo.md': 'MAIN_FOO\n',
            'docs/main/main.md': (
                '--8<-- "README.md"\n\n'
                '--8<-- "foo.md"\n\n'
                '--8<-- "test/bar.md"\n\n'
                '--8<-- "../foo.md"\n\n'
            ),
        }
    )
    def test_markdown_extension_with_relative(self, config_dir):
        for base_path, expected in {
            '!relative': '''
                <p>(Failed to read 'README.md')</p>
                <p>MAIN_FOO</p>
                <p>(Failed to read 'test/bar.md')</p>
                <p>ROOT_FOO</p>''',
            '!relative $docs_dir': '''
                <p>(Failed to read 'README.md')</p>
                <p>ROOT_FOO</p>
                <p>TEST_BAR</p>
                <p>(Failed to read '../foo.md')</p>''',
            '!relative $config_dir/docs': '''
                <p>(Failed to read 'README.md')</p>
                <p>ROOT_FOO</p>
                <p>TEST_BAR</p>
                <p>(Failed to read '../foo.md')</p>''',
            '!relative $config_dir': '''
                <p>CONFIG_README</p>
                <p>(Failed to read 'foo.md')</p>
                <p>(Failed to read 'test/bar.md')</p>
                <p>(Failed to read '../foo.md')</p>''',
        }.items():
            with self.subTest(base_path=base_path):
                cfg = f'''
                    site_name: test
                    use_directory_urls: false
                    markdown_extensions:
                      - mkdocs.tests.build_tests:
                          base_path: {base_path}
                '''
                config = base.load_config(
                    io.StringIO(cfg), config_file_path=os.path.join(config_dir, 'mkdocs.yml')
                )

                with self._assert_build_logs(''):
                    build.build(config)
                main_path = Path(config_dir, 'site', 'main', 'main.html')
                self.assertTrue(main_path.is_file())
                self.assertIn(textwrap.dedent(expected), main_path.read_text())

    # Test build.site_directory_contains_stale_files

    @tempdir(files=['index.html'])
    def test_site_dir_contains_stale_files(self, site_dir):
        self.assertTrue(build.site_directory_contains_stale_files(site_dir))

    @tempdir()
    def test_not_site_dir_contains_stale_files(self, site_dir):
        self.assertFalse(build.site_directory_contains_stale_files(site_dir))


class _TestPreprocessor(markdown.preprocessors.Preprocessor):
    def __init__(self, base_path: str) -> None:
        self.base_path = base_path

    def run(self, lines: list[str]) -> list[str]:
        for i, line in enumerate(lines):
            if m := re.search(r'^--8<-- "(.+)"$', line):
                try:
                    lines[i] = Path(self.base_path, m[1]).read_text()
                except OSError:
                    lines[i] = f"(Failed to read {m[1]!r})\n"
        return lines


class _TestExtension(markdown.extensions.Extension):
    def __init__(self, base_path: str) -> None:
        self.base_path = str(base_path)

    def extendMarkdown(self, md: markdown.Markdown) -> None:
        md.preprocessors.register(_TestPreprocessor(self.base_path), "mkdocs_test", priority=32)


makeExtension = _TestExtension



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from mkdocs.structure.nav import Section


class StructureItem(metaclass=abc.ABCMeta):
    """An item in MkDocs structure - see concrete subclasses Section, Page or Link."""

    @abc.abstractmethod
    def __init__(self):
        ...

    parent: Section | None = None
    """The immediate parent of the item in the site navigation. `None` if it's at the top level."""

    @property
    def is_top_level(self) -> bool:
        return self.parent is None

    title: str | None
    is_section: bool = False
    is_page: bool = False
    is_link: bool = False

    @property
    def ancestors(self) -> Iterable[StructureItem]:
        if self.parent is None:
            return []
        return [self.parent, *self.parent.ancestors]

    def _indent_print(self, depth: int = 0) -> str:
        return ('    ' * depth) + repr(self)



import os
import unittest

from mkdocs import exceptions
from mkdocs.config import base
from mkdocs.config import config_options as c
from mkdocs.config import defaults
from mkdocs.config.base import ValidationError
from mkdocs.tests.base import change_dir, tempdir


class ConfigBaseTests(unittest.TestCase):
    def test_unrecognised_keys(self):
        conf = defaults.MkDocsConfig()
        conf.load_dict(
            {
                'not_a_valid_config_option': "test",
            }
        )

        failed, warnings = conf.validate()

        self.assertEqual(
            warnings,
            [
                (
                    'not_a_valid_config_option',
                    'Unrecognised configuration name: not_a_valid_config_option',
                )
            ],
        )

    def test_missing_required(self):
        conf = defaults.MkDocsConfig()

        errors, warnings = conf.validate()

        self.assertEqual(
            errors, [('site_name', ValidationError('Required configuration not provided.'))]
        )
        self.assertEqual(warnings, [])

    @tempdir()
    def test_load_from_file(self, temp_dir):
        """
        Users can explicitly set the config file using the '--config' option.
        Allows users to specify a config other than the default `mkdocs.yml`.
        """
        with open(os.path.join(temp_dir, 'mkdocs.yml'), 'w') as config_file:
            config_file.write("site_name: MkDocs Test\n")
        os.mkdir(os.path.join(temp_dir, 'docs'))

        cfg = base.load_config(config_file=config_file.name)
        self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
        self.assertEqual(cfg.site_name, 'MkDocs Test')

    @tempdir()
    def test_load_default_file(self, temp_dir):
        """Test that `mkdocs.yml` will be loaded when '--config' is not set."""
        with open(os.path.join(temp_dir, 'mkdocs.yml'), 'w') as config_file:
            config_file.write("site_name: MkDocs Test\n")
        os.mkdir(os.path.join(temp_dir, 'docs'))
        with change_dir(temp_dir):
            cfg = base.load_config(config_file=None)
            self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
            self.assertEqual(cfg.site_name, 'MkDocs Test')

    @tempdir()
    def test_load_default_file_with_yaml(self, temp_dir):
        """Test that `mkdocs.yml` will be loaded when '--config' is not set."""
        with open(os.path.join(temp_dir, 'mkdocs.yaml'), 'w') as config_file:
            config_file.write("site_name: MkDocs Test\n")
        os.mkdir(os.path.join(temp_dir, 'docs'))
        with change_dir(temp_dir):
            cfg = base.load_config(config_file=None)
            self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
            self.assertEqual(cfg.site_name, 'MkDocs Test')

    @tempdir()
    def test_load_default_file_prefer_yml(self, temp_dir):
        """Test that `mkdocs.yml` will be loaded when '--config' is not set."""
        with open(os.path.join(temp_dir, 'mkdocs.yml'), 'w') as config_file1:
            config_file1.write("site_name: MkDocs Test1\n")
        with open(os.path.join(temp_dir, 'mkdocs.yaml'), 'w') as config_file2:
            config_file2.write("site_name: MkDocs Test2\n")

        os.mkdir(os.path.join(temp_dir, 'docs'))
        with change_dir(temp_dir):
            cfg = base.load_config(config_file=None)
            self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
            self.assertEqual(cfg.site_name, 'MkDocs Test1')

    def test_load_from_missing_file(self):
        with self.assertRaisesRegex(
            exceptions.ConfigurationError, "Config file 'missing_file.yml' does not exist."
        ):
            base.load_config(config_file='missing_file.yml')

    @tempdir()
    def test_load_from_open_file(self, temp_path):
        """`load_config` can accept an open file descriptor."""
        config_fname = os.path.join(temp_path, 'mkdocs.yml')
        config_file = open(config_fname, 'w+')
        config_file.write("site_name: MkDocs Test\n")
        config_file.flush()
        os.mkdir(os.path.join(temp_path, 'docs'))

        cfg = base.load_config(config_file=config_file)
        self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
        self.assertEqual(cfg.site_name, 'MkDocs Test')
        # load_config will always close the file
        self.assertTrue(config_file.closed)

    @tempdir()
    def test_load_from_closed_file(self, temp_dir):
        """
        The `serve` command with auto-reload may pass in a closed file descriptor.
        Ensure `load_config` reloads the closed file.
        """
        with open(os.path.join(temp_dir, 'mkdocs.yml'), 'w') as config_file:
            config_file.write("site_name: MkDocs Test\n")
        os.mkdir(os.path.join(temp_dir, 'docs'))

        cfg = base.load_config(config_file=config_file)
        self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
        self.assertEqual(cfg.site_name, 'MkDocs Test')

    @tempdir()
    def test_load_missing_required(self, temp_dir):
        """`site_name` is a required setting."""
        with open(os.path.join(temp_dir, 'mkdocs.yml'), 'w') as config_file:
            config_file.write("site_dir: output\nsite_url: https://www.mkdocs.org\n")
        os.mkdir(os.path.join(temp_dir, 'docs'))

        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(exceptions.Abort):
                base.load_config(config_file=config_file.name)
        self.assertEqual(
            '\n'.join(cm.output),
            "ERROR:mkdocs.config:Config value 'site_name': Required configuration not provided.",
        )

    def test_pre_validation_error(self):
        class InvalidConfigOption(c.BaseConfigOption):
            def pre_validation(self, config, key_name):
                raise ValidationError('pre_validation error')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(errors, [('invalid_option', ValidationError('pre_validation error'))])
        self.assertEqual(warnings, [])

    def test_run_validation_error(self):
        class InvalidConfigOption(c.BaseConfigOption):
            def run_validation(self, value):
                raise ValidationError('run_validation error')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(errors, [('invalid_option', ValidationError('run_validation error'))])
        self.assertEqual(warnings, [])

    def test_post_validation_error(self):
        class InvalidConfigOption(c.BaseConfigOption):
            def post_validation(self, config, key_name):
                raise ValidationError('post_validation error')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(errors, [('invalid_option', ValidationError('post_validation error'))])
        self.assertEqual(warnings, [])

    def test_pre_and_run_validation_errors(self):
        """A pre_validation error does not stop run_validation from running."""

        class InvalidConfigOption(c.BaseConfigOption):
            def pre_validation(self, config, key_name):
                raise ValidationError('pre_validation error')

            def run_validation(self, value):
                raise ValidationError('run_validation error')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(
            errors,
            [
                ('invalid_option', ValidationError('pre_validation error')),
                ('invalid_option', ValidationError('run_validation error')),
            ],
        )
        self.assertEqual(warnings, [])

    def test_run_and_post_validation_errors(self):
        """A run_validation error stops post_validation from running."""

        class InvalidConfigOption(c.BaseConfigOption):
            def run_validation(self, value):
                raise ValidationError('run_validation error')

            def post_validation(self, config, key_name):
                raise ValidationError('post_validation error')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(errors, [('invalid_option', ValidationError('run_validation error'))])
        self.assertEqual(warnings, [])

    def test_validation_warnings(self):
        class InvalidConfigOption(c.BaseConfigOption):
            def pre_validation(self, config, key_name):
                self.warnings.append('pre_validation warning')

            def run_validation(self, value):
                self.warnings.append('run_validation warning')

            def post_validation(self, config, key_name):
                self.warnings.append('post_validation warning')

        conf = base.Config(schema=(('invalid_option', InvalidConfigOption()),))

        errors, warnings = conf.validate()

        self.assertEqual(errors, [])
        self.assertEqual(
            warnings,
            [
                ('invalid_option', 'pre_validation warning'),
                ('invalid_option', 'run_validation warning'),
                ('invalid_option', 'post_validation warning'),
            ],
        )

    @tempdir()
    def test_load_from_file_with_relative_paths(self, config_dir):
        """
        When explicitly setting a config file, paths should be relative to the
        config file, not the working directory.
        """
        config_fname = os.path.join(config_dir, 'mkdocs.yml')
        with open(config_fname, 'w') as config_file:
            config_file.write("docs_dir: src\nsite_name: MkDocs Test\n")
        docs_dir = os.path.join(config_dir, 'src')
        os.mkdir(docs_dir)

        cfg = base.load_config(config_file=config_file)
        self.assertTrue(isinstance(cfg, defaults.MkDocsConfig))
        self.assertEqual(cfg.site_name, 'MkDocs Test')
        self.assertEqual(cfg.docs_dir, docs_dir)
        self.assertEqual(cfg.config_file_path, config_fname)
        self.assertIsInstance(cfg.config_file_path, str)

    def test_get_schema(self):
        class FooConfig:
            z = c.URL()
            aa = c.Type(int)

        self.assertEqual(
            base.get_schema(FooConfig),
            (
                ('z', FooConfig.z),
                ('aa', FooConfig.aa),
            ),
        )



#!/usr/bin/env python

import unittest

from mkdocs.structure.toc import get_toc
from mkdocs.tests.base import dedent, get_markdown_toc


class TableOfContentsTests(unittest.TestCase):
    def test_indented_toc(self):
        md = dedent(
            """
            # Heading 1
            ## Heading 2
            ### Heading 3
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
                Heading 2 - #heading-2
                    Heading 3 - #heading-3
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 1)

    def test_indented_toc_html(self):
        md = dedent(
            """
            # Heading 1
            ## <code>Heading</code> 2
            ## Heading 3
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
                Heading 2 - #heading-2
                Heading 3 - #heading-3
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 1)

    def test_flat_toc(self):
        md = dedent(
            """
            # Heading 1
            # Heading 2
            # Heading 3
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
            Heading 2 - #heading-2
            Heading 3 - #heading-3
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 3)

    def test_flat_h2_toc(self):
        md = dedent(
            """
            ## Heading 1
            ## Heading 2
            ## Heading 3
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
            Heading 2 - #heading-2
            Heading 3 - #heading-3
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 3)

    def test_mixed_toc(self):
        md = dedent(
            """
            # Heading 1
            ## Heading 2
            # Heading 3
            ### Heading 4
            ### Heading 5
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
                Heading 2 - #heading-2
            Heading 3 - #heading-3
                Heading 4 - #heading-4
                Heading 5 - #heading-5
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 2)

    def test_mixed_html(self):
        md = dedent(
            """
            # Heading 1
            ## Heading 2
            # Heading 3
            ### Heading 4
            ### <a>Heading 5</a>
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
                Heading 2 - #heading-2
            Heading 3 - #heading-3
                Heading 4 - #heading-4
                Heading 5 - #heading-5
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 2)

    def test_nested_anchor(self):
        md = dedent(
            """
            # Heading 1
            ## Heading 2
            # Heading 3
            ### Heading 4
            ### <a href="/">Heading 5</a>
            """
        )
        expected = dedent(
            """
            Heading 1 - #heading-1
                Heading 2 - #heading-2
            Heading 3 - #heading-3
                Heading 4 - #heading-4
                Heading 5 - #heading-5
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 2)

    def test_entityref(self):
        md = dedent(
            """
            # Heading & 1
            ## Heading > 2
            ### Heading < 3
            """
        )
        expected = dedent(
            """
            Heading &amp; 1 - #heading-1
                Heading &gt; 2 - #heading-2
                    Heading &lt; 3 - #heading-3
            """
        )
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 1)

    def test_charref(self):
        md = '# &#64;Header'
        expected = '&#64;Header - #header'
        toc = get_toc(get_markdown_toc(md))
        self.assertEqual(str(toc).strip(), expected)
        self.assertEqual(len(toc), 1)

    def test_level(self):
        md = dedent(
            """
            # Heading 1
            ## Heading 1.1
            ### Heading 1.1.1
            ### Heading 1.1.2
            ## Heading 1.2
            """
        )
        toc = get_toc(get_markdown_toc(md))

        def get_level_sequence(items):
            for item in items:
                yield item.level
                yield from get_level_sequence(item.children)

        self.assertEqual(tuple(get_level_sequence(toc)), (1, 2, 3, 3, 2))



import unittest
from textwrap import dedent

import yaml

from mkdocs.tests.base import load_config
from mkdocs.utils import templates


class UtilsTemplatesTests(unittest.TestCase):
    def test_script_tag(self):
        cfg_yaml = dedent(
            '''
            extra_javascript:
              - some_plain_javascript.js
              - implicitly_as_module.mjs
              - path: explicitly_as_module.mjs
                type: module
              - path: deferred_plain.js
                defer: true
              - path: scripts/async_module.mjs
                type: module
                async: true
              - path: 'aaaaaa/"my script".mjs'
                type: module
                async: true
                defer: true
              - path: plain.mjs
            '''
        )
        config = load_config(**yaml.safe_load(cfg_yaml))
        config.extra_javascript.append('plain_string.mjs')

        self.assertEqual(
            [
                str(templates.script_tag_filter({'page': None, 'base_url': 'here'}, item))
                for item in config.extra_javascript
            ],
            [
                '<script src="here/some_plain_javascript.js"></script>',
                '<script src="here/implicitly_as_module.mjs" type="module"></script>',
                '<script src="here/explicitly_as_module.mjs" type="module"></script>',
                '<script src="here/deferred_plain.js" defer></script>',
                '<script src="here/scripts/async_module.mjs" type="module" async></script>',
                '<script src="here/aaaaaa/&#34;my script&#34;.mjs" type="module" defer async></script>',
                '<script src="here/plain.mjs"></script>',
                '<script src="here/plain_string.mjs"></script>',
            ],
        )
